import sys
from pathlib import Path
from rich import print

from mlf_core.common.load_yaml import load_yaml_file


def load_mlf_core_template_version(handle: str, yaml_path: str) -> str:
    """
    Load the version of the template specified by the handler from the given yaml file path.

    :param handle: The template handle
    :param yaml_path: Path to the yaml file
    :return: The version number to the given handles template
    """
    available_templates = load_yaml_file(yaml_path)
    parts = handle.split('-')
    if len(parts) == 2:
        return available_templates[parts[0]][parts[1]]['version']
    elif len(parts) == 3:
        return available_templates[parts[0]][parts[1]][parts[2]]['version']


def load_project_template_version_and_handle(project_dir: Path) -> (str, str):
    """
    Load the template version when the user synced its project with mlf-core the last time.
    If no sync has been done so far, its the version of the mlf-core template the user created the project initially with.
    NOTE: This is NOT the projects current version (they are independent from each other)!!!

    :param project_dir: Top level directory of the users project.
    :return: The version number of the mlf-core template when the user created the project and the projects template handle.
    """
    project_dir = str(project_dir)
    try:
        ct_meta = load_yaml_file(f'{project_dir}/.mlf_core.yml')
        # split the template version at first space to omit the mlf-core bump-version tag and return it and the the handle
        return ct_meta['template_version'].split(" ", 1)[0], ct_meta['template_handle']
    except FileNotFoundError:
        print(f'[bold red]No .mlf_core.yml found at {project_dir}. Is this a mlf-core project?')
        sys.exit(1)



import collections
import io
from rich.console import Console
from rich import print
import click
import sys
from pathlib import Path
from configparser import NoSectionError

from mlf_core import __version__
from mlf_core.common.suggest_similar_commands import MAIN_COMMANDS
from mlf_core.common.levensthein_dist import most_similar_command
from mlf_core.bump_version.bump_version import VersionBumper


class HelpErrorHandling(click.Group):
    """
    Customise the help command
    """

    def __init__(self, name=None, commands=None, **kwargs):
        super(HelpErrorHandling, self).__init__(name, commands, **kwargs)
        self.commands = commands or collections.OrderedDict()

    def list_commands(self, ctx):
        return self.commands

    def main_options(self, ctx, formatter) -> None:
        """
        Load the main options and display them in a customized option section.
        :param ctx: clicks context
        :param formatter: the formatter for output
        """
        ct_main_options = []
        # NOTE: this only works for options as arguments do not have a help attribute per default
        for p in ctx.command.params:
            ct_main_options.append(('--' + p.name + ': ', p.help))
        ct_main_options.append(('--help: ', '   Get detailed info on a command.'))
        with formatter.section(self.get_rich_value('Options')):
            for t in ct_main_options:
                formatter.write_text(f'{t[0] + t[1]}')

    def format_help(self, ctx, formatter):
        """
        Call format_help function with mlf-core customized functions.
        """
        self.format_usage(ctx, formatter)
        self.format_options(ctx, formatter)
        self.format_commands(ctx, formatter)

    def format_usage(self, ctx, formatter):
        """
        Overwrite format_usage method of class MultiCommand for customized usage section output.
        """
        formatter.write_text(f'{self.get_rich_value("Usage:")} mlf-core {" ".join(super().collect_usage_pieces(ctx))}')

    def format_options(self, ctx, formatter):
        """
        Overwrite the format_options method of class MultiCommand for customized option output.
        This is internally called by format_help() which itself is called by get_help().
        """
        self.main_options(ctx, formatter)

    def format_commands(self, ctx, formatter):
        """
        Overwrite the format_commands method of class MultiCommand for customized commands output.
        """
        formatter.width = 120

        with formatter.section(self.get_rich_value("General Commands")):
            formatter.write_text(f"{self.commands.get('list').name}\t\t\t{self.commands.get('list').get_short_help_str(limit=150)}")
            formatter.write_text(f"{self.commands.get('info').name}\t\t\t{self.commands.get('info').get_short_help_str(limit=150)}")
            formatter.write_text(f"{self.commands.get('config').name}\t\t\t{self.commands.get('config').get_short_help_str(limit=150)}")
            formatter.write_text(f"{self.commands.get('upgrade').name}\t\t\t{self.commands.get('upgrade').get_short_help_str(limit=150)}")

        with formatter.section(self.get_rich_value("Commands for mlf-core projects")):
            formatter.write_text(f"{self.commands.get('create').name}\t\t\t{self.commands.get('create').get_short_help_str(limit=150)}")
            formatter.write_text(f"{self.commands.get('lint').name}\t\t\t{self.commands.get('lint').get_short_help_str(limit=150)}")
            formatter.write_text(f"{self.commands.get('fix-artifact-paths').name}\t{self.commands.get('fix-artifact-paths').get_short_help_str(limit=150)}")
            formatter.write_text(f"{self.commands.get('bump-version').name}\t\t{self.commands.get('bump-version').get_short_help_str(limit=150)}")
            formatter.write_text(f"{self.commands.get('sync').name}\t\t\t{self.commands.get('sync').get_short_help_str(limit=150)}")

        with formatter.section(self.get_rich_value("Examples")):
            formatter.write_text("$ mlf-core create")
            formatter.write_text("$ mlf-core bump-version 1.0.0 .")
            formatter.write_text("$ mlf-core config all")
            formatter.write_text("$ mlf-core info mlflow")

        with formatter.section(self.get_rich_value("Learn more")):
            formatter.write_text("Use mlf-core <command> --help for more information about a command. You may also want to take a look at our docs at "
                                 "https://mlf-core.readthedocs.io/.")

        with formatter.section(self.get_rich_value("Feedback")):
            formatter.write_text("We are always curious about your opinion on mlf-core. Join our Discord at "
                                 "https://discord.gg/Mv8sAcq and drop us a message.")

    def get_command(self, ctx, cmd_name):
        """
        Override the get_command of Click.
        If an unknown command is given, try to determine a similar command.
        If no similar command could´ve been found, exit with an error message.
        Else use the most similar command while printing a status message for the user.

        :param ctx: The given Click context for the group
        :param cmd_name: The command invoked with Click
        """
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        sim_commands, action = most_similar_command(cmd_name, MAIN_COMMANDS)

        matches = [cmd for cmd in self.list_commands(ctx) if cmd in sim_commands]

        # no similar commands could be found
        if not matches:
            ctx.fail(click.style('Unknown command and no similar command was found!', fg='red'))
        elif len(matches) == 1 and action == 'use':
            print(f'[bold red]Unknown command! Will use best match [green]{matches[0]}')
            return click.Group.get_command(self, ctx, matches[0])
        elif len(matches) == 1 and action == 'suggest':
            ctx.fail(click.style('Unknown command! Did you mean ', fg='red') + click.style(f'{matches[0]}?', fg='green'))

        # a few similar commands were found, print a info message
        ctx.fail(click.style('Unknown command. Most similar commands were', fg='red') + click.style(f'{", ".join(sorted(matches))}', fg='red'))

    @staticmethod
    def args_not_provided(ctx, cmd: str) -> None:
        """
        Print a fail message depending on the command.
        :param ctx: Click app context
        :param cmd: The invoked subcommand
        """
        if cmd == 'info':
            print(f'[bold red]Failed to execute [bold green]{cmd.upper()}.\n[bold blue]Please provide a valid handle like [bold green]cli '
                  '[bold blue]as argument.')
            sys.exit(1)

        elif cmd == 'bump-version':
            print(f'[bold red]Failed to execute [bold green]{cmd.upper()}.\n[bold blue]Please provide a new version like [bold green]1.2.3 '
                  '[bold blue]as first argument.')
            sys.exit(1)

        elif cmd == 'config':
            print(f'[bold red]Failed to execute [bold green]{cmd.upper()}.\n[bold blue]Please provide a valid argument. You can choose general, pat or all.')
            sys.exit(1)

    def get_rich_value(self, output: str, is_header=True) -> str:
        """
        Return a string which contains the output to console rendered by rich for the click formatter.
        :param output: the output string, that should be rendered by rich
        :param is_header: is output a header section?
        """
        sio = io.StringIO()
        console = Console(file=sio, force_terminal=True)
        if is_header:
            console.print(f"[bold #1874cd]{output}")

        return sio.getvalue().replace('\n', '')


def print_project_version(ctx, param, value) -> None:
    """
    Print the current project version.
    """
    # if context uses resilient parsing (no changes of execution flow) or no flag value is provided, do nothing
    if not value or ctx.resilient_parsing:
        return
    try:
        print(f'[bold blue]Current project version is [bold green]{VersionBumper(Path.cwd(), False).CURRENT_VERSION}!')
        ctx.exit()
    # currently, its only possible to get project version from top level project dir where the mlf-core.cfg file is
    except NoSectionError:
        ctx.fail(click.style('Unable to read from mlf-core.cfg file.\nMake sure your current working directory has a mlf-core.cfg file '
                             'when running bump-version with the --project-version flag!', fg='red'))


def print_mlfcore_version(ctx, param, value):
    """
    Print mlf-core version
    """
    # if context uses resilient parsing (no changes of execution flow) or no flag value is provided, do nothing
    if not value or ctx.resilient_parsing:
        return
    try:
        print(f'[bold blue]Mlf-core version: {__version__}')
        ctx.exit()
    except click.ClickException:
        ctx.fail('An error occurred fetching mlf-core version!')


class CustomHelpSubcommand(click.Command):
    """
    Customize the help output for each subcommand
    """

    def __init__(self, *args, **kwargs):
        super(CustomHelpSubcommand, self).__init__(*args, **kwargs)

    def format_help(self, ctx, formatter):
        """
        Custom implementation of formatting help for each subcommand.
        Use the overwritten format functions this class provides to output help for each subcommand mlf-core provides.
        """
        formatter.width = 120
        self.format_usage(ctx, formatter)
        self.format_help_text(ctx, formatter)
        self.format_options(ctx, formatter)

    def format_usage(self, ctx, formatter):
        """
        Custom implementation if formatting the usage of each subcommand.
        Usage section with a styled header will be printed.
        """
        formatter.write_text(f'{self.get_rich_value("Usage: ")}mlf-core {self.name} {" ".join(self.collect_usage_pieces(ctx))}')

    def format_help_text(self, ctx, formatter):
        """
        Custom implementation of formatting the help text of each subcommand.
        The help text will be printed as normal. A separate arguments section will be added below with all arguments and a short help message
        for each of them and a styled header in order to keep things separated.
        """
        formatter.write_paragraph()
        formatter.write_text(self.help)
        args = [('--' + param.name, param.helpmsg) for param in self.params if type(param) == CustomArg]
        if args:
            with formatter.section(self.get_rich_value("Arguments")):
                formatter.write_dl(args)

    def format_options(self, ctx, formatter):
        """
        Custom implementation of formatting the options of each subcommand.
        The options will be displayed in their relative order with their corresponding help message and a styled header.
        """
        options = [('--' + param.name.replace('_', '-'), param.help) for param in self.params if type(param) == click.core.Option]
        help_option = self.get_help_option(ctx)
        options.append(('--' + help_option.name, help_option.help))
        with formatter.section(self.get_rich_value("Options")):
            formatter.write_dl(options)

    def get_rich_value(self, output: str, is_header=True) -> str:
        """
        Return a string which contains the output to console rendered by rich for the click formatter.
        :param output: the output string, that should be rendered by rich
        :param is_header: is output a header section?
        """
        sio = io.StringIO()
        console = Console(file=sio, force_terminal=True)
        if is_header:
            console.print(f"[bold #1874cd]{output}")

        return sio.getvalue().replace('\n', '')


class CustomArg(click.Argument):
    """
    A custom argument implementation of click.Argument class in order to provide a short helpmessage for each argument of a command.
    """

    def __init__(self, *args, **kwargs):
        self.helpmsg = kwargs.pop('helpmsg')
        super().__init__(*args, **kwargs)



from mlf_core.common.suggest_similar_commands import SIMILARITY_SUGGEST_FACTOR, SIMILARITY_USE_FACTOR


def levensthein_dist(input_command: str, candidate: str) -> int:
    """
    Implement the Levenshtein distance algorithm to determine, in case of a non-existing handle,
    if theres a very similar command to suggest.

    :param input_command: The non-existing handle the user gave as input
    :param candidate: The (possible similar) alternative command
    :return: The similarity between the two strings measured by the levensthein distance
    """

    if not input_command or not candidate:
        return max(len(input_command), len(candidate))  # at least one string is empty

    dp_table = [[0 for col in range(len(input_command) + 1)] for row in range(len(candidate) + 1)]

    dp_table[0] = list(range(0, len(input_command) + 1))
    for i in range(1, len(candidate) + 1):
        dp_table[i][0] = i

    # now choose minimum levensthein distance from the three option delete/replace/insert
    # if chars are the same -> levensthein distance is the same as for those substring without these chars of input_command
    # and candidate

    for i in range(1, len(candidate) + 1):
        for j in range(1, len(input_command) + 1):
            # choose minimum edit distance from delete, replace or insert at current substring
            if input_command[j - 1] == candidate[i - 1]:
                dp_table[i][j] = dp_table[i - 1][j - 1]
            else:
                dp_table[i][j] = min(min(dp_table[i][j - 1], dp_table[i - 1][j - 1]), dp_table[i - 1][j]) + 1

    return dp_table[len(candidate)][len(input_command)]


def most_similar_command(command: str, command_list: set) -> (list, str):
    """
    Determine whether its possible to suggest a similar command.
    The similarity is determined by the levensthein distance and a factor (currently 1/3)
    sets a limit where a similar command is useful to be automatically used. If the difference diff is 1/3 < diff <= 2/3, one
    or more similar commands could be suggested, but not used automatically.

    :param command_list: The commands that are available by the users specific action
    :param command: The command given by the user
    :return: A list of similar command(s) or the empty string if there's none and a string that indicates the action to be taken
    """
    min_use = 999999  # some random large integer -> we will never have handles that are larger than 1000 character
    min_suggest = 999999
    sim_command_use = []
    sim_command_suggest = []

    # for each valid handle calculate the levensthein distance and if one is found that is a new minimal distance,
    # replace it and take this handle as the most similar command.
    for handle in command_list:
        dist = levensthein_dist(command, handle)

        # the more restrict condition for automatic use
        lim_use = int(len(command) * SIMILARITY_USE_FACTOR)

        # the weaker condition for command suggestion
        lim_suggest = int(len(command) * SIMILARITY_SUGGEST_FACTOR)

        # check if the command is close to the inputted command so it can be automatically used
        if lim_use >= dist:
            if min_use > dist:  # and min >= dist:
                min_use = dist
                sim_command_use = [handle]
            elif min_use == dist:
                sim_command_use.append(handle)

        # the input is not very close to any command, but maybe a similar one can be suggested?
        elif lim_use < dist <= lim_suggest:
            if min_suggest > dist:  # and min >= dist:
                min_suggest = dist
                sim_command_suggest = [handle]
            elif min_suggest == dist:
                sim_command_suggest.append(handle)

    # return the use list, as those are closer, but if its empty, return the list of suggested commands (or if that is empty too, an empty list)
    return (sim_command_use, 'use') if sim_command_use else (sim_command_suggest, 'suggest') if sim_command_suggest else ([], '')



import os

import requests
from pkg_resources import parse_version

from mlf_core.lint.template_linter import TemplateLinter, GetLintingFunctionsMeta, files_exist_linting


class PackagePredictionLint(TemplateLinter, metaclass=GetLintingFunctionsMeta):
    def __init__(self, path):
        super().__init__(path)

    def lint(self):
        super().lint_project(self, self.methods)

    def check_dependencies_not_outdated(self) -> bool:
        """
        Check that every dependency from project's requirements.txt is the latest version available at PyPi.

        :return Bool flag that shows code execution went right (used for testing purposes)
        """

        def check_dependencies(filename: str) -> None:
            """
            Check for a given file whether no dependencies are outdated.
            :param filename: Name of the dependency file to parse (either requirements.txt or requirements_dev.txt)
            """
            with open(f'{self.path}/{filename}') as req_file:
                dependencies = [line[:-1].split('==') for line in req_file]
            for dependency in dependencies:
                if len(dependency) == 2:
                    _check_pip_package(dependency[0], dependency[1])

        def _check_pip_package(pip_dependency_name, pip_dependency_version) -> None:
            """
            Query PyPi package information.
            Sends a HTTP GET request to the PyPi remote API.

            :param pip_dependency_name: The name of the dependency
            :param pip_dependency_version: Dependency version used by the user's project
            """
            pip_api_url = f'https://pypi.python.org/pypi/{pip_dependency_name}/json'
            try:
                response = requests.get(pip_api_url, timeout=10)
            except requests.exceptions.Timeout:
                self.warned.append(('cli-python-2', f'PyPi API timed out: {pip_api_url}'))
            except requests.exceptions.ConnectionError:
                self.warned.append(('cli-python-2', f'PyPi API Connection error: {pip_api_url}'))
            else:
                if response.status_code == 200:
                    pip_dep_json = response.json()
                    latest_dependency_version = pip_dep_json['info']['version']
                    if parse_version(pip_dependency_version) < parse_version(latest_dependency_version):
                        self.warned.append(('package-prediction-2', f'Version {pip_dependency_version} of {pip_dependency_name} is not the latest available: '
                        f'{latest_dependency_version}'))  # noqa: E128
                else:
                    self.failed.append(('package-prediction-2',
                                        f'Could not find pip dependency using the PyPi API: {pip_dependency_name}=={pip_dependency_version}'))

        # check general dependencies
        check_dependencies('requirements.txt')
        # check development dependencies
        check_dependencies('requirements_dev.txt')
        return True

    def python_files_exist(self) -> None:
        """
        Checks a given project directory for required files.
        Iterates through the templates's directory content and checkmarks files for presence.
        Files that **must** be present::
            'setup.py',
            'setup.cfg',
            'MANIFEST.in',
        Files that *should* be present::
            '.github/workflows/build_package.yml',
            '.github/workflows/publish_package.yml',
            '.github/workflows/flake8.yml',
        Files that *must not* be present::
            none
        Files that *should not* be present::
            '__pycache__'
        """

        # NB: Should all be files, not directories
        # List of lists. Passes if any of the files in the sublist are found.
        files_fail = [
            ['setup.py'],
            ['setup.cfg'],
            ['MANIFEST.in'],
        ]
        files_warn = [
            [os.path.join('.github', 'workflows', 'build_package.yml')],
            [os.path.join('.github', 'workflows', 'publish_package.yml')],
            [os.path.join('.github', 'workflows', 'run_flake8_linting.yml')],
        ]

        # List of strings. Fails / warns if any of the strings exist.
        files_fail_ifexists = [
            '__pycache__'
        ]
        files_warn_ifexists = [

        ]

        files_exist_linting(self, files_fail, files_fail_ifexists, files_warn, files_warn_ifexists, handle='package-prediction')



from collections.abc import MutableMapping
from contextlib import suppress


def delete_keys_from_dict(dictionary: MutableMapping, keys: list) -> None:
    """
    Deletes all key instances in an arbitrarily nested dictionary inplace

    :param dictionary: dictionary of which the keys are deleted
    :param keys: list of keys to delete
    """
    for key in keys:
        with suppress(KeyError):
            del dictionary[key]
    for value in dictionary.values():
        if isinstance(value, MutableMapping):
            delete_keys_from_dict(value, keys)


def is_nested_dictionary(dictionary: dict) -> bool:
    """
    Determines whether a dictionary is nested or not

    :param dictionary: dictionary to examine
    :return: True if dictionary is nested, false otherwise
    """
    return any(isinstance(_, dict) for _ in dictionary.values())



import os
from pathlib import Path
from pkg_resources import parse_version

from mlf_core.common.load_yaml import load_yaml_file
from mlf_core.lint.template_linter import TemplateLinter, files_exist_linting, GetLintingFunctionsMeta

CWD = os.getcwd()


class MlflowPytorchLint(TemplateLinter, metaclass=GetLintingFunctionsMeta):
    def __init__(self, path):
        super().__init__(path)

    def lint(self):
        super().lint_project(self, self.methods)

    def pytorch_files_exist(self) -> None:
        """
        Checks a given project directory for required files.
        Iterates through the templates's directory content and checkmarks files for presence.
        Files that **must** be present::
            'MLproject',
            'environment.yml',
            'project_slug/'
        Files that *should* be present::
            '.github/workflows/train_cpu.yml',
            '.github/workflows/lint.yml',
        Files that *must not* be present::
            none
        Files that *should not* be present::
            '__pycache__'
        """

        # NB: Should all be files, not directories
        # List of lists. Passes if any of the files in the sublist are found.
        files_fail = [
            ['MLproject'],
            ['environment.yml'],
            [f'{self.project_slug_no_hyphen}/mlf_core/mlf_core.py']
        ]

        files_warn = [
            [os.path.join('.github', 'workflows', 'train_cpu.yml')],
            [os.path.join('.github', 'workflows', 'lint.yml')],
        ]

        # List of strings. Fails / warns if any of the strings exist.
        files_fail_ifexists = [
            '__pycache__'
        ]
        files_warn_ifexists = [

        ]

        files_exist_linting(self, files_fail, files_fail_ifexists, files_warn, files_warn_ifexists, handle='mlflow-pytorch')

    def pytorch_reproducibility_seeds(self) -> None:
        """
        Verifies that all CPU and GPU reproducibility settings for Pytorch are enabled
        Required are:
             trainer.deterministic = True,
             trainer.benchmark = False,
             set_general_random_seeds(general_seed),
             set_pytorch_random_seeds(pytorch_seed, num_of_gpus)
        """
        passed_pytorch_reproducibility_seeds = True
        passed_pytorch_reproducibility_seeds_mlf_core = True
        entry_point_file_path = f'{self.path}/{self.project_slug_no_hyphen}/{self.project_slug_no_hyphen}.py'
        mlf_core_file_path = f'{self.path}/{self.project_slug_no_hyphen}/mlf_core/mlf_core.py'
        with open(entry_point_file_path) as f:
            project_slug_entry_point_content = set(map(lambda line: line.strip(), f.readlines()))
        with open(mlf_core_file_path) as f:
            mlf_core_content = set(map(lambda line: line.strip(), f.readlines()))

        expected_lines_pytorch_reproducibility = ['trainer.deterministic = True',
                                                  'trainer.benchmark = False']

        expected_lines_pytorch_mlf_core = ['torch.manual_seed(seed)',
                                           'torch.set_deterministic(True)',
                                           'torch.cuda.manual_seed(seed)',
                                           'torch.cuda.manual_seed_all(seed)  # For multiGPU']

        # check for existence of all required entry point files lines for reproducibility
        for expected_line in expected_lines_pytorch_reproducibility:
            if expected_line not in project_slug_entry_point_content:
                passed_pytorch_reproducibility_seeds = False
                self.failed.append(('mlflow-pytorch-2', f'{expected_line} not found in {entry_point_file_path}'))

        # check for existence of required mlf_core lines
        for expected_line in expected_lines_pytorch_mlf_core:
            if expected_line not in mlf_core_content:
                passed_pytorch_reproducibility_seeds_mlf_core = False
                self.failed.append(('mlflow-pytorch-3', f'{expected_line} not found in {entry_point_file_path}'))

        # only pass linting when all seeds and mlf_core settings are set
        if passed_pytorch_reproducibility_seeds and passed_pytorch_reproducibility_seeds_mlf_core:
            self.passed.append(('mlflow-pytorch-2', 'All required reproducibility settings enabled.'))
            self.passed.append(('mlflow-pytorch-3', 'All required reproducibility mlf_core settings enabled.'))

    def pytorch_no_atomic_operations(self) -> None:
        """
        Verifies that the project does not use any of the potentially non-deterministic atomicAdd functions.

        There are some PyTorch functions that use CUDA functions that can be a source of nondeterminism.
        One class of such CUDA functions are atomic operations, in particular atomicAdd, which can lead to the order of additions being nondetermnistic.
        Because floating-point addition is not perfectly associative for floating-point operands,
        atomicAdd with floating-point operands can introduce different floating-point rounding errors on each evaluation,
        which introduces a source of nondeterministic variance (aka noise) in the result.
        PyTorch functions that use atomicAdd in the forward kernels include torch.Tensor.index_add_(), torch.Tensor.scatter_add_(), torch.bincount().
        A number of operations have backwards kernels that use atomicAdd, including torch.nn.functional.embedding_bag(),
        torch.nn.functional.ctc_loss(), torch.nn.functional.interpolate(), and many forms of pooling, padding, and sampling.
        There is currently no simple way of avoiding nondeterminism in these functions.
        Additionally, the backward path for repeat_interleave() operates nondeterministically on the CUDA backend because repeat_interleave() is implemented
        using index_select(), the backward path for which is implemented using index_add_(), which is known to operate
        nondeterministically (in the forward direction) on the CUDA backend (see above).

        Source: https://pytorch.org/docs/stable/notes/randomness.html
        """
        atomic_add_functions = [
            'index_add',
            'index_select',
            'scatter_add',
            'bincount',
            'embedding_bag',
            'interpolate',
            'repeat_interleave',
            'histc'
            'AvgPool3d',
            'AdaptiveAvgPool2d',
            'AdaptiveAvgPool3d',
            'MaxPool3d',
            'AdaptiveMaxPool2d',
            'FractionalMaxPool2d',
            'FractionalMaxPool3d',
            'ReflectionPad1d',
            'ReflectionPad2d',
            'ReplicationPad1d',
            'ReplicationPad2d'
            'ReplicationPad3d',
            'NLLLoss',
            'CTCLoss',
            'EmbeddingBag'

            # interpolate when called on a CUDA tensor that requires grad and one of the following modes is used: - linear - bilinear - bicubic - trilinear
            # TODO COOKIETEMPLE: This should ideally be solved with a nice regex, which would also help when people have several parameters
            'interpolate(\'linear\')',
            'interpolate(\"linear\")',
            'interpolate(mode=\'linear\')',
            'interpolate(mode=\"linear\")',

            'interpolate(\'bilinear\')',
            'interpolate(\"bilinear\")',
            'interpolate(mode=\'bilinear\')',
            'interpolate(mode=\"bilinear\")',

            'interpolate(bicubic\'bicubic\')',
            'interpolate(\"bicubic\")',
            'interpolate(mode=\'bicubic\')',
            'interpolate(mode=\"bicubic\")',

            'interpolate(bicubic\'trilinear\')',
            'interpolate(\"trilinear\")',
            'interpolate(mode=\'trilinear\')',
            'interpolate(mode=\"trilinear\")',
        ]

        verify_method_not_present(self, atomic_add_functions, 'mlflow-pytorch-3')


class MlflowTensorflowLint(TemplateLinter, metaclass=GetLintingFunctionsMeta):
    def __init__(self, path):
        super().__init__(path)

    def lint(self):
        super().lint_project(self, self.methods)

    def tensorflow_files_exist(self) -> None:
        """
        Checks a given project directory for required files.
        Iterates through the templates's directory content and checkmarks files for presence.
        Files that **must** be present::
            'MLproject,
            'environment.yml',
            'project_slug_no_hyphen/mlf_core/mlf_core.py',
        Files that *should* be present::
            '.github/workflows/train_cpu.yml',
            '.github/workflows/lint.yml',
        Files that *must not* be present::
            none
        Files that *should not* be present::
            '__pycache__'
        """

        # NB: Should all be files, not directories
        # List of lists. Passes if any of the files in the sublist are found.
        files_fail = [
            ['MLproject'],
            ['environment.yml'],
            [f'{self.project_slug_no_hyphen}/mlf_core/mlf_core.py']
        ]
        files_warn = [
            [os.path.join('.github', 'workflows', 'train_cpu.yml')],
            [os.path.join('.github', 'workflows', 'lint.yml')],
        ]

        # List of strings. Fails / warns if any of the strings exist.
        files_fail_ifexists = [
            '__pycache__'
        ]
        files_warn_ifexists = [

        ]

        files_exist_linting(self, files_fail, files_fail_ifexists, files_warn, files_warn_ifexists, handle='mlflow-tensorflow')

    def tensorflow_reproducibility_seeds(self) -> None:
        """
        Verifies that all CPU and GPU reproducibility settings for Tensorflow are enabled
        Required are:
        set_general_random_seeds(dict_args["general_seed"])
        set_tensorflow_random_seeds(dict_args["tensorflow_seed"])
        def set_tensorflow_random_seeds(seed):
            tf.random.set_seed(seed)
            tf.config.threading.set_intra_op_parallelism_threads = 1  # CPU only -> https://github.com/NVIDIA/tensorflow-determinism
            tf.config.threading.set_inter_op_parallelism_threads = 1  # CPU only
            os.environ['TF_DETERMINISTIC_OPS'] = '1'
        """
        passed_tensorflow_reproducibility_seeds = True
        entry_point_file_path = f'{self.path}/{self.project_slug_no_hyphen}/{self.project_slug_no_hyphen}.py'
        with open(entry_point_file_path) as f:
            project_slug_entry_point_content = list(map(lambda line: line.strip(), f.readlines()))

        expected_lines_tensorflow_reproducibility = ['MLFCore.set_general_random_seeds(dict_args["general_seed"])',
                                                     'MLFCore.set_tensorflow_random_seeds(dict_args["tensorflow_seed"])']

        for expected_line in expected_lines_tensorflow_reproducibility:
            if expected_line not in project_slug_entry_point_content:
                passed_tensorflow_reproducibility_seeds = False
                self.failed.append(('mlflow-tensorflow-2', f'{expected_line} not found in {entry_point_file_path}'))

        if passed_tensorflow_reproducibility_seeds:
            self.passed.append(('mlflow-tensorflow-2', 'All required reproducibility settings enabled.'))

        # TODO COOKIETEMPLE: Investigate whether there is a reasonable way of linting for
        # train_dataset = mnist_train.map(scale).cache().shuffle(buffer_size, seed=tensorflow_seed, reshuffle_each_iteration=False).batch(BATCH_SIZE)

    def tensorflow_non_deterministic_functions(self) -> None:
        """
        Verifies that no non-deterministic functions of Tensorflow are used.
        """
        non_deterministic_tf_functions = ['softmax_cross_entropy_with_logits',
                                          'sparse_softmax_cross_entropy_with_logits']

        verify_method_not_present(self, non_deterministic_tf_functions, 'mlflow-tensorflow-3')


class MlflowXGBoostLint(TemplateLinter, metaclass=GetLintingFunctionsMeta):
    def __init__(self, path):
        super().__init__(path)

    def lint(self):
        super().lint_project(self, self.methods)

    def xgboost_files_exist(self) -> None:
        """
        Checks a given project directory for required files.
        Iterates through the templates's directory content and checkmarks files for presence.
        Files that **must** be present::
            'MLproject',
            'environment.yml',
            'project_slug/mlf_core/mlf_core.py'
        Files that *should* be present::
            '.github/workflows/train_cpu.yml',
            '.github/workflows/lint.yml',
        Files that *must not* be present::
            none
        Files that *should not* be present::
            '__pycache__'
        """

        # NB: Should all be files, not directories
        # List of lists. Passes if any of the files in the sublist are found.
        files_fail = [
            ['MLproject'],
            ['environment.yml'],
            [f'{self.project_slug_no_hyphen}/mlf_core/mlf_core.py']
        ]
        files_warn = [
            [os.path.join('.github', 'workflows', 'train_cpu.yml')],
            [os.path.join('.github', 'workflows', 'lint.yml')],
        ]

        # List of strings. Fails / warns if any of the strings exist.
        files_fail_ifexists = [
            '__pycache__'
        ]
        files_warn_ifexists = [

        ]

        files_exist_linting(self, files_fail, files_fail_ifexists, files_warn, files_warn_ifexists, handle='mlflow-xgboost')

    def xgboost_reproducibility_seeds(self) -> None:
        """
        Verifies that all CPU and GPU reproducibility settings for XGBoost are enabled
        """
        passed_xgboost_reproducibility_seeds = True
        entry_point_file_path = f'{self.path}/{self.project_slug_no_hyphen}/{self.project_slug_no_hyphen}.py'
        with open(entry_point_file_path) as f:
            project_slug_entry_point_content = list(map(lambda line: line.strip(), f.readlines()))

        expected_lines_xgboost_reproducibility = ['MLFCore.set_general_random_seeds(dict_args["general_seed"])',
                                                  'MLFCore.set_xgboost_random_seeds(dict_args["xgboost_seed"], param)']

        for expected_line in expected_lines_xgboost_reproducibility:
            if expected_line not in project_slug_entry_point_content:
                passed_xgboost_reproducibility_seeds = False
                self.failed.append(('mlflow-xgboost-2', f'{expected_line} not found in {entry_point_file_path}'))

        if passed_xgboost_reproducibility_seeds:
            self.passed.append(('mlflow-xgboost-2', 'All required reproducibility settings enabled.'))

    def xgboost_version(self) -> None:
        """
        Verifies that the XGBoost version is at least 1.1.0, since reproducibility cannot be guaranteed elsewise.
        """
        # Verify that XGBoost version is greater than 1.1.0
        conda_env = load_yaml_file(f'{self.path}/environment.yml')
        conda_only = list(filter(lambda dep: '::' in dep, conda_env['dependencies']))
        pip_only = list(filter(lambda dep: isinstance(dep, dict), conda_env['dependencies']))[0]['pip']

        for dependency in conda_only:
            if 'xgboost' in dependency:
                split = dependency.split('==')
                current_version = parse_version(split[-1])
                if current_version < parse_version('1.1.0'):
                    self.failed.append(('mlflow-xgboost-3',
                                        f'XGBoost version {current_version} is not at least 1.1.0. Reproducibility cannot be guaranteed.'))

        for dependency in pip_only:
            if 'xgboost' in dependency:
                split = dependency.split('==')
                current_version = parse_version(split[-1])
                if current_version < parse_version('1.1.0'):
                    self.failed.append(('mlflow-xgboost-3',
                                        f'XGBoost version {current_version} is not at least 1.1.0. Reproducibility cannot be guaranteed.'))

    def xgboost_no_all_reduce(self) -> None:
        """
        Verifies that all_reduce is not used.
        https://github.com/dmlc/xgboost/issues/5023
        """
        all_reduce_functions = ['all_reduce']
        verify_method_not_present(self, all_reduce_functions, 'mlflow-xgboost-4')


def verify_method_not_present(calling_class: TemplateLinter, functions_to_check: list, linting_code: str):
    """
    Checks all Python files for a list of strings/functions, which should not be present.
    :param calling_class: The TemplateLinter subclass. Required to append the found errors correctly.
    :param functions_to_check: The list of functions, which are not allowed to be present in any Python file
    :param linting_code: A linting code build from the handle and the error code e.g. mlflow-pytorch-1
    """
    # We should only expect those functions in all *.py files, so only read those
    for root, dirs, files in os.walk(f'{calling_class.path}/{calling_class.project_slug_no_hyphen}'):
        for file in files:
            if file.endswith(".py"):
                file_to_check_full_path = Path(root) / file
                with open(file_to_check_full_path) as f:
                    content_stripped = list(map(lambda line: line.strip(), f.readlines()))
                    for function in functions_to_check:
                        for line_code in content_stripped:
                            if function in line_code:
                                calling_class.failed.append((linting_code,
                                                             f'{function} found in {file_to_check_full_path} operates non-deterministically.'))



import os
import sys
import logging
import requests
import json
from base64 import b64encode
from nacl import encoding, public
from pathlib import Path
from cryptography.fernet import Fernet
from distutils.dir_util import copy_tree
from subprocess import Popen, PIPE
from github import Github, GithubException
from git import Repo, exc
from ruamel.yaml import YAML
from rich import print
from collections import OrderedDict

from mlf_core.create.domains.mlf_core_template_struct import MlfcoreTemplateStruct
from mlf_core.custom_cli.questionary import mlf_core_questionary_or_dot_mlf_core
from mlf_core.common.load_yaml import load_yaml_file
from mlf_core.config.config import ConfigCommand

log = logging.getLogger(__name__)


def create_push_github_repository(project_path: str, creator_ctx: MlfcoreTemplateStruct, tmp_repo_path: str) -> None:
    """
    Creates a Github repository for the created template and pushes the template to it.
    Prompts the user for the required specifications.

    :param creator_ctx: Full Template Struct. Github username may be updated if an organization repository is warranted.
    :param project_path: The path to the recently created project
    :param tmp_repo_path: Path to the empty cloned repo
    """
    try:
        if not is_git_accessible():
            return

        # the personal access token for GitHub
        access_token = handle_pat_authentification()

        # Login to Github
        log.debug('Logging into Github.')
        print('[bold blue]Logging into Github')
        authenticated_github_user = Github(access_token)
        user = authenticated_github_user.get_user()

        # Create new repository
        print('[bold blue]Creating Github repository')
        if creator_ctx.is_github_orga:
            log.debug(f'Creating a new Github repository for organizaton: {creator_ctx.github_orga}.')
            org = authenticated_github_user.get_organization(creator_ctx.github_orga)
            repo = org.create_repo(creator_ctx.project_slug, description=creator_ctx.project_short_description, private=creator_ctx.is_repo_private)
            creator_ctx.github_username = creator_ctx.github_orga
        else:
            log.debug(f'Creating a new Github repository for user: {creator_ctx.github_username}.')
            repo = user.create_repo(creator_ctx.project_slug, description=creator_ctx.project_short_description, private=creator_ctx.is_repo_private)

        print('[bold blue]Creating labels and default Github settings')
        create_github_labels(repo=repo, labels=[('DEPENDABOT', '1BB0CE')])

        repository = f'{tmp_repo_path}'

        # NOTE: github_username is the organizations name, if an organization repository is to be created

        # create the repos sync secret
        print('[bold blue]Creating repository sync secret')
        create_sync_secret(creator_ctx.github_username, creator_ctx.project_slug, access_token)

        # git clone
        log.debug(f'Cloning repository {creator_ctx.github_username}/{creator_ctx.project_slug}')
        print('[bold blue]Cloning empty Github repository')
        Repo.clone_from(f'https://{creator_ctx.github_username}:{access_token}@github.com/{creator_ctx.github_username}/{creator_ctx.project_slug}', repository)

        # Copy files which should be included in the initial commit -> basically the template
        log.debug('Copying files from the template into the cloned repository.')
        copy_tree(f'{repository}', project_path)

        # the created project repository with the copied .git directory
        cloned_repo = Repo(path=project_path)

        # git add
        log.debug('git add')
        print('[bold blue]Staging template')
        cloned_repo.git.add(A=True)

        # git commit
        log.debug('git commit')
        cloned_repo.index.commit(f'Created {creator_ctx.project_slug} with {creator_ctx.template_handle} '
                                 f'template of version {creator_ctx.template_version} using mlf-core.')

        # get the default branch of the repository as default branch since GitHub repositories are configurable by the user and can be set to any branch name
        # but mlf-core needs to know which one is the default branch in order to push to the correct remote branch and rename local branch, if necessary
        headers = {'Authorization': f'token {access_token}'}
        url = f"https://api.github.com/repos/{creator_ctx.github_username}/{creator_ctx.project_slug}"
        response = requests.get(url, headers=headers).json()
        default_branch = response['default_branch']
        log.debug(f'git push origin {default_branch}')
        print(f'[bold blue]Pushing template to Github origin {default_branch}')
        if default_branch != 'master':
            cloned_repo.git.branch('-M', f'{default_branch}')
        cloned_repo.remotes.origin.push(refspec=f'{default_branch}:{default_branch}')

        # set branch protection (all WF must pass, dismiss stale PR reviews) only when repo is public
        log.debug('Set branch protection rules.')
        if not creator_ctx.is_repo_private and not creator_ctx.is_github_orga:
            main_branch = authenticated_github_user.get_user().get_repo(name=creator_ctx.project_slug).get_branch(f'{default_branch}')
            main_branch.edit_protection(dismiss_stale_reviews=True)
        else:
            print('[bold yellow]Cannot set branch protection rules due to your repository being private or an organization repository!\n'
                  'You can set it manually later on.')

        # git create development branch
        log.debug('git checkout -b development')
        print('[bold blue]Creating development branch.')
        cloned_repo.git.checkout('-b', 'development')

        # git push to origin development
        log.debug('git push origin development')
        print('[bold blue]Pushing template to Github origin development.')
        cloned_repo.remotes.origin.push(refspec='development:development')

        # git create TEMPLATE branch
        log.debug('git checkout -b TEMPLATE')
        print('[bold blue]Creating TEMPLATE branch.')
        cloned_repo.git.checkout('-b', 'TEMPLATE')

        # git push to origin TEMPLATE
        log.debug('git push origin TEMPLATE')
        print('[bold blue]Pushing template to Github origin TEMPLATE.')
        cloned_repo.remotes.origin.push(refspec='TEMPLATE:TEMPLATE')

        # finally, checkout to development branch
        log.debug('git checkout development')
        print('[bold blue]Checking out development branch.')
        cloned_repo.git.checkout('development')

        print(f'[bold green]Successfully created a Github repository at '
              f'https://github.com/{creator_ctx.github_username}/{creator_ctx.project_slug}')

        print('[bold blue]Do not forget to make your Docker container Github Package public!')
        print('[bold blue]Read https://bit.ly/2VDwMKw for instructions.')

    except (GithubException, ConnectionError) as e:
        handle_failed_github_repo_creation(e)


def handle_pat_authentification() -> str:
    """
    Try to read the encrypted Personal Access Token for GitHub.
    If this fails (maybe there was no generated key before) notify user to config its credentials for mlf-core.

    :return: The decrypted PAT
    """
    log.debug(f'Attempting to read the personal access token from {ConfigCommand.CONF_FILE_PATH}')
    # check if the key and encrypted PAT already exist
    if os.path.exists(ConfigCommand.CONF_FILE_PATH):
        path = Path(ConfigCommand.CONF_FILE_PATH)
        yaml = YAML(typ='safe')
        settings = yaml.load(path)
        if os.path.exists(ConfigCommand.KEY_PAT_FILE) and 'pat' in settings:
            pat = decrypt_pat()
            return pat
        else:
            log.debug(f'Unable to read the personal access token from {ConfigCommand.CONF_FILE_PATH}')
            print('[bold red]Could not find encrypted personal access token!\n')
            print('[bold blue]Please navigate to Github -> Your profile -> Settings -> Developer Settings -> Personal access token -> Generate a new Token')
            print('[bold blue]Only tick \'repo\'. The token is a hidden input to mlf-core and stored encrypted locally on your machine.')
            print('[bold blue]For more information please read '
                  'https://help.github.com/en/github/authenticating-to-github/creating-a-personal-access-token-for-the-command-line\n\n')
            print('[bold blue]Lets move on to set your personal access token for your mlf-core project!')
            # set the PAT
            ConfigCommand.config_pat()
            # if the user wants to create a GitHub repo but accidentally presses no on PAT config prompt
            if not os.path.exists(ConfigCommand.KEY_PAT_FILE):
                print('[bold red]No Github personal access token found. Please set it using [green]mlf-core config github')
                sys.exit(1)
            else:
                pat = decrypt_pat()
            return pat
    else:
        print('[bold red]Cannot find a mlf-core config file! Did you delete it?')


def prompt_github_repo(dot_mlf_core: OrderedDict or None) -> (bool, bool, bool, str):
    """
    Ask user for all settings needed in order to create and push automatically to GitHub repo.

    :param dot_mlf_core: .mlf_core.yml content if passed
    :return if is git repo, if repo should be private, if user is an organization and if so, the organizations name
    """
    # if dot_mlf_core dict was passed -> return the Github related properties and do NOT prompt for them
    try:
        if dot_mlf_core:
            if not dot_mlf_core['is_github_orga']:
                return dot_mlf_core['is_github_repo'], dot_mlf_core['is_repo_private'], False, ''
            else:
                return dot_mlf_core['is_github_repo'], dot_mlf_core['is_repo_private'], dot_mlf_core['is_github_orga'], \
                       dot_mlf_core['github_orga']
    except KeyError:
        print('[bold red]Missing required Github properties in .mlf_core.yml file!')

    # No dot_mlf_core_dict was passed -> prompt whether to create a Github repository and the required settings
    create_git_repo, private, is_github_org, github_org = False, False, False, ''
    print('[bold blue]Automatically creating a Github repository with mlf-core is strongly recommended. '
          'Otherwise you will not be able to use all of mlf-core\'s features!\n')
    if mlf_core_questionary_or_dot_mlf_core(function='confirm',
                                            question='Do you want to create a Github repository and push your template to it?',
                                            default='Yes'):
        create_git_repo = True
        is_github_org = mlf_core_questionary_or_dot_mlf_core(function='confirm',
                                                             question='Do you want to create an organization repository?',
                                                             default='No')
        github_org = mlf_core_questionary_or_dot_mlf_core(function='text',
                                                          question='Please enter the name of the Github organization',
                                                          default='mlf-core') if is_github_org else ''
        private = mlf_core_questionary_or_dot_mlf_core(function='confirm',
                                                       question='Do you want your repository to be private?',
                                                       default='No')

    return create_git_repo, private, is_github_org, github_org


def create_sync_secret(username: str, repo_name: str, token: str) -> None:
    """
    Create the secret mlf-core uses to sync repos. The secret contains the personal access token with the repo scope.
    Following steps are required (PAT MUST have at least repo access):
    1.) Get the repos public key (and its ID) which is needed for secret's value (PAT) encryption; for private repos especially we need an authentification
        header for a successful request.
    2.) Encrypt the secret value using PyNacl (a Python binding for Javascripts LibSodium) and send the data with an authentification header (PAT) and the
        public key's ID via PUT to the Github API.

    :param username: The users github username
    :param repo_name: The repositories name
    :param token: The PAT of the user with repo scope
    """
    public_key_dict = get_repo_public_key(username, repo_name, token)
    try:
        create_secret(username, repo_name, token, public_key_dict['key'], public_key_dict['key_id'])
    except KeyError:
        print('[bold red]Could not create sync secret for repository due to GitHubs API currently unavailable. See '
              'https://mlf-core.readthedocs.io/en/latest/faq.html for more!')


def get_repo_public_key(username: str, repo_name: str, token: str) -> dict:
    """
    Get the public key for a repository via the Github API. At least for private repos, a personal access token (PAT) with the repo scope is required.

    :param username: The users github username
    :param repo_name: The repositories name
    :param token: The PAT of the user with repo scope
    :return: A dict containing the public key and its ID
    """
    query_url = f'https://api.github.com/repos/{username}/{repo_name}/actions/secrets/public-key'
    headers = {'Authorization': f'token {token}'}
    r = requests.get(query_url, headers=headers)

    return r.json()


def create_secret(username: str, repo_name: str, token: str, public_key_value: str, public_key_id: str) -> None:
    """
    Create the secret named CT_SYNC_TOKEN using a PUT request via the Github API. This request needs a PAT with the repo scope for authentification purposes.
    Using PyNacl, a Python binding for Javascripts LibSodium, it encrypts the secret value, which is required by the Github API.

    :param username: The user's github username
    :param repo_name: The repositories name
    :param token: The PAT of the user with repo scope
    :param public_key_value: The public keys value (the key) of the repos public key PyNacl uses for encryption of the secrets value
    :param public_key_id: The ID of the public key used for encryption
    """
    log.debug('Creating Github repository secret.')
    encrypted_value = encrypt_sync_secret(public_key_value, token)
    # the parameters required by the Github API
    params = {
        "encrypted_value": encrypted_value,
        "key_id": public_key_id
    }
    # the authentication header
    headers = {'Authorization': f'token {token}'}
    # the url used for PUT
    put_url = f'https://api.github.com/repos/{username}/{repo_name}/actions/secrets/MLF_CORE_SYNC_TOKEN'
    requests.put(put_url, headers=headers, data=json.dumps(params))


def encrypt_sync_secret(public_key: str, token: str) -> str:
    """
    Encrypt the sync secret (which is the PAT).

    :param public_key: Public key of the repo we want to create a secret for
    :param token: The users PAT with repo scope as the secret
    :return: The encrypted secret (PAT)
    """
    log.debug('Encrypting Github repository secret.')
    public_key = public.PublicKey(public_key.encode("utf-8"), encoding.Base64Encoder())
    sealed_box = public.SealedBox(public_key)
    encrypted = sealed_box.encrypt(token.encode("utf-8"))

    return b64encode(encrypted).decode("utf-8")


def decrypt_pat() -> str:
    """
    Decrypt the encrypted PAT.

    :return: The decrypted Personal Access Token for GitHub
    """
    log.debug(f'Decrypting personal access token using key saved in {ConfigCommand.KEY_PAT_FILE}.')
    # read key and encrypted PAT from files
    with open(ConfigCommand.KEY_PAT_FILE, 'rb') as f:
        key = f.readline()
    fer = Fernet(key)
    log.debug(f'Reading personal access token from {ConfigCommand.CONF_FILE_PATH}.')
    encrypted_pat = load_yaml_file(ConfigCommand.CONF_FILE_PATH)['pat']
    # decrypt the PAT and decode it to string
    print('[bold blue]Decrypting personal access token.')
    decrypted_pat = fer.decrypt(encrypted_pat).decode('utf-8')

    return decrypted_pat


def load_github_username() -> str:
    """
    Load the username from cfg file.

    :return: The users Github account name
    """
    return load_yaml_file(ConfigCommand.CONF_FILE_PATH)['github_username']


def handle_failed_github_repo_creation(e: ConnectionError or GithubException) -> None:
    """
    Called, when the automatic GitHub repo creation process failed during the create process. As this may have various issue sources,
    try to provide the user a detailed error message for the individual exception and inform them about what they should/can do next.

    :param e: The exception that has been thrown
    """
    # output the error dict thrown by PyGitHub due to an error related to GitHub
    if isinstance(e, GithubException):
        print('[bold red]\nError while trying to create a Github repo due to an error related to Github API. See below output for detailed information!\n')
        format_github_exception(e.data)
    # output an error that might occur due to a missing internet connection
    elif isinstance(e, ConnectionError):
        print('[bold red]Error while trying to establish a connection to https://github.com. Do you have an active internet connection?')


def format_github_exception(data: dict) -> None:
    """
    Format the github exception thrown by PyGitHub in a nice way and output it.

    :param data: The exceptions data as a dict
    """
    for section, description in data.items():
        if not isinstance(description, list):
            print(f'[bold red]{section.capitalize()}: {description}')
        else:
            print(f'[bold red]{section.upper()}: ')
            messages = [
                val if not isinstance(val, dict) and not isinstance(val, set) else github_exception_dict_repr(val) for
                val in description]
            print('[bold red]\n'.join(msg for msg in messages))


def github_exception_dict_repr(messages: dict) -> str:
    """
    String representation for Github exception dict thrown by PyGitHub.

    :param messages: The messages as a dict
    """
    return '\n'.join(f'    {section.capitalize()}: {description}' for section, description in messages.items())


def is_git_accessible() -> bool:
    """
    Verifies that git is accessible and in the PATH.

    :return: True if accessible, false if not
    """
    log.debug('Testing whether git is accessible.')
    git_installed = Popen(['git', '--version'], stdout=PIPE, stderr=PIPE, universal_newlines=True)
    (git_installed_stdout, git_installed_stderr) = git_installed.communicate()
    if git_installed.returncode != 0:
        print('[bold red]Could not find \'git\' in the PATH. Is it installed?')
        print('[bold red]Run command was: \'git --version \'')
        log.debug('git is not accessible!')
        return False

    return True


def create_github_labels(repo, labels: list) -> None:
    """
    Create github labels and add them to the repository.
    If failed, print error message.

    :param repo: The repository where the label needs to be added
    :param labels: A list of the new labels to be added
    """
    for label in labels:
        log.debug(f'Creating Github label {label[0]}')
        try:
            repo.create_label(name=label[0], color=label[1])
        except GithubException:
            log.debug(f'Unable to create label {label[0]}')
            print(f'[bold red]Unable to create label {label[0]} due to permissions')


def is_git_repo(path: Path) -> bool:
    """
    Check if directory is a git repo

    :param path: The directory to check
    :return: true if path is git repo false otherwise
    """
    try:
        _ = Repo(path).git_dir
        return True
    except exc.InvalidGitRepositoryError:
        return False



import os
import sys
import re
import logging
from packaging import version
from configparser import ConfigParser
from tempfile import mkstemp
from shutil import move, copymode
from os import fdopen, remove
from pathlib import Path
from git import Repo
from datetime import datetime
from rich import print

from mlf_core.create.github_support import is_git_repo
from mlf_core.lint.template_linter import TemplateLinter
from mlf_core.custom_cli.questionary import mlf_core_questionary_or_dot_mlf_core

log = logging.getLogger(__name__)


class VersionBumper:
    """
    Responsible for bumping the version across a mlf-core project
    """

    def __init__(self, project_dir, downgrade):
        self.parser = ConfigParser()
        self.parser.read(f'{project_dir}/mlf_core.cfg')
        self.CURRENT_VERSION = self.parser.get('bumpversion', 'current_version')
        self.downgrade_mode = downgrade
        self.top_level_dir = project_dir

    def bump_template_version(self, new_version: str, project_dir: Path) -> None:
        """
        Update the version number for all files that are whitelisted in the config file.

        INFO on valid versions: All versions must match the format like 1.0.0 or 1.1.0-SNAPSHOT; these are the only valid
        version formats mlf-core allows. A valid version therefore contains a three digits (in the range from 0 to however large it will grow)
        separated by two dots.
        Optional is the -SNAPSHOT at the end (for JVM templates especially). NOTE that versions like 1.2.3.4 or 1.2 WILL NOT be recognized as valid versions as
        well as no substring of them will be recognized.

        :param new_version: The new version number that should replace the old one in a mlf-core project
        :param project_dir: The default value is the current working directory, so we´re initially assuming the user
                             bumps the version from the projects top level directory. If this is not the case this parameter
                             shows the path where the projects top level directory is and bumps the version there
        """
        log.debug(f'Current version: {self.CURRENT_VERSION} --- New version: {new_version}')
        sections = ['bumpversion_files_whitelisted', 'bumpversion_files_blacklisted']

        # if project_dir was given as handle use cwd since we need it for git add
        ct_cfg_path = f'{str(project_dir)}/mlf_core.cfg' if str(project_dir).startswith(str(Path.cwd())) else \
            f'{str(Path.cwd())}/{project_dir}/mlf_core.cfg'
        # path to CHANGELOG.rst file
        changelog_path = f'{str(project_dir)}/CHANGELOG.rst' if str(project_dir).startswith(str(Path.cwd())) else \
            f'{str(Path.cwd())}/{project_dir}/CHANGELOG.rst'

        # keep path of all files that were changed during bump version
        changed_files = [ct_cfg_path, changelog_path]

        print(f'[bold blue]Changing version number.\nCurrent version is {self.CURRENT_VERSION}.'
              f'\nNew version will be {new_version}\n')

        # for each section (whitelisted and blacklisted files) bump the version (if allowed)
        for section in sections:
            log.debug(f'Bumping files of section: {section}.')
            for file, path in self.parser.items(section):
                not_changed, file_path = self.replace(f'{project_dir}/{path}', new_version, section)
                # only add file if the version(s) in the file were bumped
                if not not_changed:
                    path_changed = file_path if file_path.startswith(str(Path.cwd())) else f'{str(Path.cwd())}/{file_path}'
                    changed_files.append(path_changed)

        # update new version in mlf_core.cfg file
        log.debug('Updating version in mlf_core.cfg file.')
        self.parser.set('bumpversion', 'current_version', new_version)
        with open(f'{project_dir}/mlf_core.cfg', 'w') as configfile:
            self.parser.write(configfile)

        # add a new changelog section when downgrade mode is disabled
        self.add_changelog_section(new_version)

        # check if a project is a git repository and if so, commit bumped version changes
        if is_git_repo(project_dir):
            repo = Repo(project_dir)

            # git add
            print('[bold blue]Staging template')
            repo.git.add(changed_files)

            # git commit
            print('[bold blue]Committing changes to local git repository.')
            repo.index.commit(f'Bump version from {self.CURRENT_VERSION} to {new_version}')

    def replace(self, file_path: str, subst: str, section: str) -> (bool, str):
        """
        Replace a version with the new version unless the line is explicitly excluded (marked with
        <<MLF-CORE_NO_BUMP>>).
        Or, in case of blacklisted files, it ignores all lines with version numbers unless they´re explicitly marked
        for bump with tag <<MLF-CORE_FORCE_BUMP>>.

        :param file_path: The path of the file where the version should be updated
        :param subst: The new version that replaces the old one
        :param section: The current section (whitelisted or blacklisted files)

        :return: Whether a file has been changed during bumped and the path of changed file
        """
        # flag that indicates whether no changes were made inside a file
        file_is_unchanged = True
        path_changed = ''

        # Create temp file
        fh, abs_path = mkstemp()
        with fdopen(fh, 'w') as new_file:
            with open(file_path) as old_file:
                for line in old_file:
                    # update version if tags were found (and were in the right section)
                    if ('<<MLF-CORE_NO_BUMP>>' not in line and not section == 'bumpversion_files_blacklisted') or '<<MLF-CORE_FORCE_BUMP>>' in line:
                        # for info on this regex, see bump_template docstring above
                        tmp = re.sub(r'(?<!\.)\d+(?:\.\d+){2}(?:-SNAPSHOT)?(?!\.)', subst, line)
                        new_file.write(tmp)
                        if tmp != line:
                            if file_is_unchanged:
                                print(f'[bold blue]Updating version number in {file_path}')
                                file_is_unchanged = False
                                path_changed = file_path
                            print(f'[bold red]- {line.strip().replace("<!-- <<MLF-CORE_FORCE_BUMP>> -->", "")}\n'
                                  + f'[bold green]+ {tmp.strip().replace("<!-- <<MLF-CORE_FORCE_BUMP>> -->", "")}')
                            print()
                    else:
                        new_file.write(line)
        # Copy the file permissions from the old file to the new file
        copymode(file_path, abs_path)
        # Remove original file
        remove(file_path)
        # Move new file
        move(abs_path, file_path)

        return file_is_unchanged, path_changed

    def can_run_bump_version(self, new_version: str, project_dir: str) -> bool:
        """
        Ensure that all requirements are met, so that the bump version command can be run successfully.
        This included the following requirements:
        1.) The new version number matches the format like 1.1.0 or 1.1.0-SNAPSHOT required by mlf-core versions
        2.) The new version is greater than the current one
        3.) The project is a mlf-core project

        :param new_version: The new version
        :param project_dir: The directory of the project
        :return: True if bump version can be run, false otherwise.
        """
        # ensure that the entered version number matches correct format like 1.1.0 or 1.1.0-SNAPSHOT but not 1.2 or 1.2.3.4
        if not re.match(r'(?<!\.)\d+(?:\.\d+){2}((?!.)|-SNAPSHOT)(?!.)', new_version):
            print('[bold red]Invalid version specified!\nEnsure your version number has the form '
                  'of 0.0.0 or 15.100.239-SNAPSHOT')
            return False

        # ensure the version is bumped within a project created by mlf-core
        elif not Path(f'{project_dir}/mlf_core.cfg').is_file():
            print('[bold red]Did not find a mlf_core.cfg file. Make sure you are in the right directory '
                  'or specify the path to your projects bump_version.cfg file')
            return False

        # equal versions won't be accepted for bump-version
        elif new_version == self.CURRENT_VERSION:
            print(f'[bold red]The new version {new_version} cannot be equal to the current version {self.CURRENT_VERSION}.')
            return False

        # only allow bump from a SNAPSHOT version to its correspondence with -SNAPSHOT removed (like 1.0.0-SNAPSHOT to 1.0.0 but not 2.0.0)
        elif self.CURRENT_VERSION.endswith('-SNAPSHOT') and not self.CURRENT_VERSION.split('-')[0] == new_version:
            print(f'[bold red]Cannot bump {self.CURRENT_VERSION} to {new_version}.' +
                  f'[blue]\n{self.CURRENT_VERSION} as a SNAPSHOT version can only be bumped to its non-snapshot equivalent '
                  f'{self.CURRENT_VERSION.split("-")[0]}.')
            return False

        # ensure the new version is greater than the current one, if not the user wants to explicitly downgrade it
        elif not self.downgrade_mode:
            current_version_r = self.CURRENT_VERSION.replace('-SNAPSHOT', '')
            new_version_r = new_version.replace('-SNAPSHOT', '')

            # bump from x.x.x to x.x.x-SNAPSHOT should be only allowed when using the downgrade flag
            if new_version.endswith('-SNAPSHOT') and self.CURRENT_VERSION == new_version.split('-')[0]:
                print(f'[bold red]Cannot downgrade {self.CURRENT_VERSION} to its version SNAPSHOT {new_version}.' +
                      f'[blue]\nUse the -d flag if you want to downgrade {self.CURRENT_VERSION} to its SNAPSHOT version.')
                return False

            # when the current version and the new version are equal, but one is a -SNAPSHOT version return true
            elif version.parse(current_version_r) == version.parse(new_version_r) and ('-SNAPSHOT' in self.CURRENT_VERSION or '-SNAPSHOT' in new_version):
                return True

            # else check if the new version is greater than the current version
            elif version.parse(current_version_r) < version.parse(new_version_r):
                return True

            # the new version is not greater than the current one
            print(f'[bold red]The new version {new_version} is not greater than the current version {self.CURRENT_VERSION}.'
                  f'\nThe new version must be greater than the old one.')
            return False

        return True

    def check_bump_range(self, current_version: str, new_version: str) -> bool:
        """
        Check if the new version seems to be a reasonable bump or not (ignored when using the downgrade flag).
        This should not break the bump-version process, but it requires confirmation of the user.

        :param current_version: The current version
        :param new_version: The new version
        :return: If it´s a reasonable bump
        """
        cur_v_split = current_version.split('.')
        new_v_split = new_version.split('.')

        # major update like bumping from 1.8.3 to 2.0.0
        if new_v_split[0] != cur_v_split[0]:
            log.debug('Identified major version bump')
            return new_v_split[1] == '0' and new_v_split[2] == '0' and (int(new_v_split[0]) - int(cur_v_split[0]) == 1)

        # minor update like bumping from 1.8.5 to 1.9.0
        elif new_v_split[1] != cur_v_split[1]:
            log.debug('Identified minor version bump')
            return new_v_split[0] == cur_v_split[0] and new_v_split[2] == '0' and (int(new_v_split[1]) - int(cur_v_split[1]) == 1)

        # x-minor update like bumping from 1.8.5 to 1.8.6
        elif new_v_split[2] != cur_v_split[2]:
            log.debug('Identified patch version bump')
            return new_v_split[0] == cur_v_split[0] and new_v_split[1] == cur_v_split[1] and (int(new_v_split[2]) - int(cur_v_split[2]) == 1)

        # case when we bumping like 3.0.0-SNAPSHOT to 3.0.0
        log.debug('Identified SNAPSHOT version bump')
        return True

    def lint_before_bump(self) -> None:
        """
        Lint the changelog prior to bumping. Linting consists of two major points (beside checking if a CHANGELOG.rst file even exists at top level directory).

        1. Lint CHANGELOG.rst to ensure that bump-version can safely add a new section
        2. Check, whether all versions are consistent over the project
        """
        changelog_linter = TemplateLinter(path=self.top_level_dir)
        changelog_path = os.path.join(self.top_level_dir, 'CHANGELOG.rst')
        # ensure changelog exists, else abort
        if not os.path.exists(changelog_path):
            print(f'[bold red]No file named CHANGELOG.rst found at {self.top_level_dir}. Aborting!')
            sys.exit(1)
        # lint changelog and check version consistency
        log.debug('Linting changelog')
        changelog_linter.lint_changelog()
        log.debug('Linting version consistent')
        changelog_linter.check_version_consistent()
        print()
        changelog_linter._print_results()
        print()
        # if any failed linting tests, ask user for confirmation of proceeding with bump (which results in undefined behavior)
        if len(changelog_linter.failed) > 0 or len(changelog_linter.warned) > 0:
            # ask for confirmation if the user really wants to proceed bumping when linting failed
            print('[bold red]Changelog linting and/or version check failed!\nYou can fix them and try bumping again. Proceeding bump will result in '
                  'undefined behavior!')
            if not mlf_core_questionary_or_dot_mlf_core(function='confirm',
                                                        question='Do you really want to continue?',
                                                        default='n'):
                sys.exit(1)

    def add_changelog_section(self, new_version: str) -> None:
        """
        Each version bump will add a new section template to the CHANGELOG.rst
        :param new_version: The new version
        """
        log.debug('Adding new changelog section.')
        if self.downgrade_mode:
            print('[bold yellow]WARNING: Running bump-version in downgrade mode will not add a new changelog section currently!')
        else:
            date = datetime.today().strftime("%Y-%m-%d")
            # replace the SNAPSHOT SECTION header with its non-snapshot correlate
            if self.CURRENT_VERSION.endswith('-SNAPSHOT'):
                self.replace_snapshot_header(f'{self.top_level_dir}/CHANGELOG.rst', new_version, date)

            else:
                # the section template for a new changelog section
                nl = '\n'
                section = f'{new_version} ({date}){nl}{"-" * (len(new_version) + len(date) + 3)}{nl}{nl}' \
                          f'{f"**{nl}{nl}".join(["**Added", "**Fixed", "**Dependencies", "**Deprecated**"])}'

                self.insert_latest_version_section(old_changelog_file=f'{self.top_level_dir}/CHANGELOG.rst', section=section)

    def replace_snapshot_header(self, source_file_path: str, new_version: str, date: str) -> None:
        """
        Replace the SNAPSHOT header section in CHANGELOG. The pattern (currently) cannot include any newline characters, therefore no multiline support!
        :param source_file_path: Path to source file (the path where CHANGELOG lies)
        :param new_version: The new version
        :param date: Current date
        """
        log.debug('Replacing the changelog header in the changelog file.')
        # create a temp file (requires to be explicitly deleted later)
        fh, target_file_path = mkstemp()
        # read from old file (the source file) and write into new file (the target file)
        with open(target_file_path, 'w') as target_file:
            with open(source_file_path, 'r') as source_file:
                for line in source_file:
                    pattern, subst = '', ''
                    # check if the line is a header section with SNAPSHOT version
                    if re.match(r'^(?<!\.)\d+(?:\.\d+){2}(?!\.)-SNAPSHOT \(\d\d\d\d-\d\d-\d\d\)$', line):
                        dotted_snapshot_line = source_file.readline()
                        next_new_line = source_file.readline()  # noqa: F841 necessary to omit an additional newline
                        snapshot_date = line.split('(')[1][:-2]  # extract date of SNAPSHOT version adding
                        pattern = f'{self.CURRENT_VERSION} ({snapshot_date})'
                        subst = f'{new_version} ({date})\n{(len(new_version) + len(date) + 3) * "-"}'
                        # replace -SNASPHOT in the header and adjust the dotted line below to the new header length
                        target_file.write(line.replace(pattern, subst))
                        target_file.write(dotted_snapshot_line.replace('-', ''))
                    else:
                        # else just write the line to the new file
                        target_file.write(line.replace(pattern, subst))
        # remove old file
        remove(source_file_path)
        # move new file to replace old file
        move(target_file_path, source_file_path)

    def insert_latest_version_section(self, old_changelog_file: str, section: str) -> None:
        """
        Insert the new changelog section as the latest section right after the header
        :param old_changelog_file: path to the current CHANGELOG.rst file
        :param section: the new section template block for changelog
        """
        log.debug('Inserting latest version section into the changelog.')
        # create a temp file (requires to be explicitly deleted later)
        fh, target_file_path = mkstemp()
        # read from old file (the source file) and write into new file (the target file)
        with open(target_file_path, 'w') as target_file:
            with open(old_changelog_file, 'r') as source_file:
                for line in source_file:
                    # check if the line is the header section with the latest version
                    if re.match(rf'^{self.CURRENT_VERSION} \(\d\d\d\d-\d\d-\d\d\)$', line):
                        target_file.write(f'{section}\n\n\n')
                    target_file.write(line)
        # remove old file
        remove(old_changelog_file)
        # move new file to replace old file
        move(target_file_path, old_changelog_file)



import io
import os
import re
import logging
import configparser
import requests
import rich.progress
import rich.markdown
import rich.panel
import rich.console
from pkg_resources import parse_version
from rich import print
from packaging import version
from itertools import groupby

from mlf_core.common.load_yaml import load_yaml_file
from mlf_core.util.dir_util import pf, find_filepath_in_dir
from mlf_core.util.rich import console

log = logging.getLogger(__name__)


class TemplateLinter(object):
    """Object to hold linting information and results.
    Attributes:
        files (list): A list of files found sduring the linting process.
        path (str): Path to the project directory.
        failed (list): A list of tuples of the form: `(<error no>, <reason>)`
        passed (list): A list of tuples of the form: `(<passed no>, <reason>)`
        warned (list): A list of tuples of the form: `(<warned no>, <reason>)`
    """

    def __init__(self, path='.'):
        self.path = path
        self.files = []
        self.project_slug = load_yaml_file(f'{path}/.mlf_core.yml')['project_slug']
        self.project_slug_no_hyphen = load_yaml_file(f'{path}/.mlf_core.yml')['project_slug_no_hyphen']
        self.passed = []
        self.warned = []
        self.failed = []

    def lint_project(self, calling_class, check_functions: list = None, is_subclass_calling=True) -> None:
        """Main linting function.
        Takes the template directory as the primary input and iterates through
        the different linting checks in order. Collects any warnings or errors
        and returns summary at completion. Raises an exception if there is a
        critical error that makes the rest of the tests pointless (eg. no
        project script). Results from this function are printed by the main script.

        :param calling_class: The class that calls the function -> used to get the class methods, which are the linting methods
        :param check_functions: List of functions of the calling class that should be checked. If not set, the default TemplateLinter check functions are called
        :param is_subclass_calling: Indicates whether a domain specific linter calls the linting or not
        """
        # Called on its own, so not from a subclass -> run general linting
        if check_functions is None:
            # Fetch all general linting functions
            check_functions = [func for func in dir(TemplateLinter) if (callable(getattr(TemplateLinter, func)) and not func.startswith('_'))]
            # Remove internal functions
            check_functions = list(set(check_functions).difference({'lint_project'}))
            log.debug(f'Linting functions of general linting are:\n {check_functions}')
            # Remove mlflow specific linting functions if not applicable
            if 'mlflow' not in self.__class__.__name__.lower():
                check_functions = list(filter(lambda func: not func.startswith('mlflow'), check_functions))

        progress = rich.progress.Progress(
            "[bold green]{task.description}",
            rich.progress.BarColumn(bar_width=None),
            "[bold yellow]{task.completed} of {task.total}[reset] [bold green]{task.fields[func_name]}",
        )
        with progress:
            lint_progress = progress.add_task('Running lint checks', total=len(check_functions), func_name=check_functions)
            for fun_name in check_functions:
                log.debug(f'Running linting function: {fun_name}')
                progress.update(lint_progress, advance=1, func_name=fun_name)
                if fun_name == 'check_files_exist':
                    getattr(calling_class, fun_name)(is_subclass_calling)
                else:
                    getattr(calling_class, fun_name)()

    def check_files_exist(self, is_subclass_calling=True):
        """Checks a given project directory for required files.
        Iterates through the project's directory content and checkmarks files
        for presence.
        Files that **must** be present::
            'Dockerfile',
            'mlf_core.cfg'
            'Makefile'
            'README.rst'
            'CHANGELOG.rst'
            '[LICENSE, LICENSE.md, LICENCE, LICENCE.md]'
            'docs/index.rst'
            'docs/readme.rst'
            'docs/changelog.rst'
            'docs/installation.rst'
            'docs/usage.rst'
        Files that *should* be present::
            '.gitignore',
            '.dependabot/config.yml',
            '.github/ISSUE_TEMPLATE/bug_report.md',
            '.github/ISSUE_TEMPLATE/general_question.md',
            '.github/ISSUE_TEMPLATE/feature_request.md',
            '.github/pull_request.md',
        Files that *must not* be present::
            none
        Files that *should not* be present::
            none
        Raises:
            An AssertionError if .mlf_core.yml is not found found.
        """

        # NB: Should all be files, not directories
        # List of lists. Passes if any of the files in the sublist are found.
        files_fail = [
            ['Dockerfile'],
            ['mlf_core.cfg'],
            ['README.rst'],
            ['CHANGELOG.rst'],
            ['LICENSE', 'LICENSE.md', 'LICENCE', 'LICENCE.md'],  # NB: British / American spelling
            [os.path.join('docs', 'index.rst')],
            [os.path.join('docs', 'readme.rst')],
            [os.path.join('docs', 'changelog.rst')],
            [os.path.join('docs', 'usage.rst')],
        ]

        files_warn = [
            ['.gitignore'],
            [os.path.join('.github', 'ISSUE_TEMPLATE', 'bug_report.md')],
            [os.path.join('.github', 'ISSUE_TEMPLATE', 'feature_request.md')],
            [os.path.join('.github', 'ISSUE_TEMPLATE', 'general_question.md')],
            [os.path.join('.github', 'pull_request_template.md')]
        ]

        # List of strings. Fails / warns if any of the strings exist.
        files_fail_ifexists = [

        ]

        files_warn_ifexists = [

        ]

        # First - critical files. Check that this is actually a mlf-core based project
        if not os.path.isfile(pf(self, '.mlf_core.yml')):
            raise AssertionError('.mlf_core.yml not found! Is this a mlf-core project?')

        files_exist_linting(self, files_fail, files_fail_ifexists, files_warn, files_warn_ifexists, is_subclass_calling)

    def lint_changelog(self):
        """
        Lint the Changelog.rst file
        """
        changelog_path = os.path.join(self.path, 'CHANGELOG.rst')
        linter = ChangelogLinter(changelog_path, self)
        if not len(linter.changelog_content) < 3:
            # lint header first
            header_lint_code, header_detected, header_lint_passed = linter.lint_header()
            if header_lint_code != -1 and header_detected:
                # if head linting did not found any errors lint sections
                section_lint_passed = linter.lint_changelog_section()
                if section_lint_passed and header_lint_passed:
                    self.passed.append(('general-6', 'Changelog linting passed!'))
            elif not header_detected:
                self.failed.append(('general-6', 'Changelog does not seem to contain a header or your header syntax is wrong!'))
        else:
            self.failed.append(('general-6', 'Changelog does not seem to contain a header and/or at least one section!'))

    def check_docker(self):
        """
        Checks that Dockerfile contains the string ``FROM``
        """
        fn = os.path.join(self.path, 'Dockerfile')
        with open(fn, 'r') as fh:
            content = fh.read()

        # Implicitly also checks if empty.
        if 'FROM' in content:
            self.passed.append(('general-2', 'Dockerfile check passed'))
            return

        self.failed.append((2, 'Dockerfile check failed'))

    def check_mlf_core_todos(self) -> None:
        """
        Go through all template files looking for the string 'TODO MLF-CORE:' or 'MLF-CORE TODO:'
        """
        ignore = ['.git']
        if os.path.isfile(os.path.join(self.path, '.gitignore')):
            with io.open(os.path.join(self.path, '.gitignore'), 'rt', encoding='latin1') as file:
                for line in file:
                    ignore.append(os.path.basename(line.strip().rstrip('/')))
        for root, dirs, files in os.walk(self.path):
            # Ignore files
            for ignore_file in ignore:
                if ignore_file in dirs:
                    dirs.remove(ignore_file)
                if ignore_file in files:
                    files.remove(ignore_file)
            for fname in files:
                with io.open(os.path.join(root, fname), 'rt', encoding='latin1') as file:
                    for line in file:
                        if any(todostring in line for todostring in ['TODO MLF-CORE:', 'MLF-CORE TODO:']):
                            line = line.replace('<!--', '') \
                                .replace('-->', '') \
                                .replace('# TODO MLF-CORE: ', '') \
                                .replace('// TODO MLF-CORE: ', '') \
                                .replace('TODO MLF-CORE: ', '') \
                                .replace('# MLF-CORE TODO: ', '') \
                                .replace('// MLF-CORE TODO: ', '') \
                                .replace('MLF-CORE TODO: ', '') \
                                .strip()
                            self.warned.append(('general-3', f'TODO string found in {self._wrap_quotes(fname)}: {line}'))

    def check_no_cookiecutter_strings(self) -> None:
        """
        Verifies that no cookiecutter strings are in any of the files
        """
        for root, dirs, files in os.walk(self.path):
            for fname in files:
                with io.open(os.path.join(root, fname), 'rt', encoding='latin1') as file:
                    if file.name.endswith('.pyc'):
                        continue
                    for line in file:
                        regex = re.compile(r'{\s?.* cookiecutter.*\s?}')  # noqa W605
                        if regex.search(line):
                            line = f'{line[:50 - len(fname)]}..'
                            self.warned.append(('general-4', f'Cookiecutter string found in \'{fname}\': {line}'))

    def check_version_consistent(self) -> None:
        """
        This method should check that the version is consistent across all files.
        """
        parser = configparser.ConfigParser()
        parser.read(f'{self.path}/mlf_core.cfg')
        sections = ['bumpversion_files_whitelisted', 'bumpversion_files_blacklisted']

        current_version = parser.get('bumpversion', 'current_version')

        cwd = os.getcwd()
        os.chdir(self.path)

        # check if the version matches current version in each listed file (depending on whitelisted or blacklisted)
        for section in sections:
            for file, path in parser.items(section):
                self._check_version_match(path, current_version, section)
        os.chdir(cwd)
        # Pass message if there weren't any inconsistencies within the version numbers
        if not any('general-5' in tup[0] for tup in self.failed):
            self.passed.append(('general-5', 'Versions were consistent over all files'))

    def _check_version_match(self, path: str, version: str, section: str) -> None:
        """
        Check if the versions in a file are consistent with the current version in the mlf_core.cfg
        :param path: The current file-path to check
        :param version: The current version of the project specified in the mlf_core.cfg file
        :param section: The current section (blacklisted or whitelisted files)
        """
        with open(path) as file:
            for line in file:
                # if a tag is found and (depending on whether its a white or blacklisted file) check if the versions are matching
                if ('<<MLF-CORE_NO_BUMP>>' not in line and not section == 'bumpversion_files_blacklisted') or '<<MLF-CORE_FORCE_BUMP>>' in line:
                    line_version = re.search(r'(?<!\.)\d+(?:\.\d+){2}(?:-SNAPSHOT)?(?!\.)', line)
                    if line_version:
                        line_version = line_version.group(0)
                        # No match between the current version number and version in source code file
                        if line_version != version:
                            corrected_line = re.sub(r'(?<!\.)\d+(?:\.\d+){2}(?:-SNAPSHOT)?(?!\.)', version, line)
                            self.failed.append(('general-5', f'Version number don´t match in\n {path}: \n {line.strip()} should be {corrected_line.strip()}'))

    def mlflow_check_conda_environment(self) -> None:
        passed_conda_check = True
        conda_env = load_yaml_file(f'{self.path}/environment.yml')

        # Verify that the structure is somewhat reasonable
        for section in ['name', 'channels', 'dependencies']:
            if section not in conda_env:
                passed_conda_check = False
                self.failed.append(('general-7', f'Section {section} missing from environment.yml file!'))

        conda_only = list(filter(lambda dep: '::' in dep, conda_env['dependencies']))

        # Verify that all Conda dependencies have a pinned version number and are up to date
        for idx, dependency in enumerate(conda_only):
            comment_token = conda_env['dependencies'].ca.items.get(idx)
            check_dependency = True
            if comment_token and 'MLF-CORE IGNORE' in comment_token[0].value:
                check_dependency = False
            else:
                # after an = sign there should be a specified version
                split = dependency.split('==') if '==' in dependency else dependency.split('=')
                if len(split) < 2:
                    passed_conda_check = False
                    check_dependency = False
                    self.failed.append(('general-7', f'Dependency {dependency} does not have a pinned version!'))
            if check_dependency:
                self._check_anaconda_package(dependency, conda_env)

        # all pip dependencies are inside a dict
        pip_only = list(filter(lambda dep: isinstance(dep, dict), conda_env['dependencies']))[0]['pip']

        # Verify that all PyPI/pip dependencies have a pinned version number and are up to date
        for idx, dependency in enumerate(pip_only):
            comment_token = pip_only.ca.items.get(idx)
            check_dependency = True
            if comment_token and 'MLF-CORE IGNORE' in comment_token[0].value:
                check_dependency = False
            else:
                # after an = sign there should be a specified version
                split = dependency.split('==') if '==' in dependency else dependency.split('=')
                if len(split) < 2:
                    passed_conda_check = False
                    check_dependency = False
                    self.failed.append(('general-7', f'Dependency {dependency} does not have a pinned version!'))
            if check_dependency:
                self._check_pip_package(dependency)

        if passed_conda_check:
            self.passed.append(('general-7', 'Passed conda environment checks.'))

    def _check_anaconda_package(self, dependency, conda_environment):
        """Query conda package information.
        Sends a HTTP GET request to the Anaconda remote API.
        :param dependency: A conda package name.
        :param conda_environment: A dictionary of the read in environment.yml
        """
        # Check if each dependency is the latest available version
        dependency_split = dependency.split('==') if '==' in dependency else dependency.split('=')
        dependency_name, dependency_version = dependency_split[0], dependency_split[1]
        dep_channels = conda_environment.get('channels', [])

        if '::' in dependency_name:
            dep_channels = [dependency_name.split('::')[0]]
            dependency_name = dependency_name.split('::')[1]

        # 'defaults' isn't actually a channel name. See https://docs.anaconda.com/anaconda/user-guide/tasks/using-repositories/
        if 'defaults' in dep_channels:
            dep_channels.remove('defaults')
            dep_channels.extend(['main', 'anaconda', 'r', 'free', 'archive', 'anaconda-extras'])
        for ch in dep_channels:
            anaconda_api_url = f'https://api.anaconda.org/package/{ch}/{dependency_name}'
            try:
                response = requests.get(anaconda_api_url, timeout=10)
            except requests.exceptions.Timeout:
                self.warned.append(('general-7', f'Anaconda API timed out: {anaconda_api_url}'))
            except requests.exceptions.ConnectionError:
                self.warned.append(('general-7', 'Could not connect to Anaconda API'))
            else:
                if response.status_code == 200:
                    conda_dep_json = response.json()
                    latest_dependency_version = conda_dep_json['latest_version']
                    latest_dependency_version = latest_dependency_version.split('=')[0]
                    dependency_version = dependency_version.split('==')[0]
                    if parse_version(dependency_version) < parse_version(latest_dependency_version):
                        self.warned.append(('general-7', f'Version {dependency_version} of {dependency_name}'
                                                         f' is not the latest available: {latest_dependency_version}'))
                    return
                elif response.status_code != 404:
                    self.warned.append(('general-7', f'Anaconda API returned unexpected response code `{response.status_code}`'
                                                     f' for: {anaconda_api_url}\n{response}'))
                elif response.status_code == 404:
                    print(f'[red]Could not find {dependency} in conda channel {ch}')
        else:
            # We have looped through each channel and had a 404 response code on everything
            self.failed.append(('general-7', f'Could not find Conda dependency {dependency} using the Anaconda API'))

    def _check_pip_package(self, dependency):
        """Query PyPi package information.
        Sends a HTTP GET request to the PyPi remote API.
        :param dependency: A PyPi package name.
        """
        pip_dependency_name, pip_dependency_version = dependency.split('==')
        pip_api_url = "https://pypi.python.org/pypi/{}/json".format(pip_dependency_name)
        try:
            response = requests.get(pip_api_url, timeout=10)
        except requests.exceptions.Timeout:
            self.warned.append(('general-7', f'PyPi API timed out: {pip_api_url}'))
        except requests.exceptions.ConnectionError:
            self.warned.append(('general-7', f'PyPi API Connection error: {pip_api_url}'))
        else:
            if response.status_code == 200:
                pip_dep_json = response.json()
                latest_dependency_version = pip_dep_json['info']['version']
                if parse_version(pip_dependency_version) < parse_version(latest_dependency_version):
                    self.warned.append(('general-7', f'Version {pip_dependency_version} of {pip_dependency_name}'
                                                     f' is not the latest available: {latest_dependency_version}'))
            else:
                self.failed.append(('general-7', f'Could not find pip dependency using the PyPi API: {dependency}'))

    def mlflow_mlf_core_py_complete(self) -> None:
        """
        Verifies that the random_seed method exists and

        """
        mlf_core_py_path = find_filepath_in_dir('mlf_core.py', self.path)
        if not mlf_core_py_path:
            self.failed.append(('mlflow-general-8', 'mlf_core.py could not be found!'))
            return

        with open(mlf_core_py_path) as f:
            mlf_core_py_content = list(map(lambda line: line.strip(), f.readlines()))

        # Verify that general_random_seeds is complete
        expected_lines_general_random_seeds = ['def set_general_random_seeds(seed):',
                                               'os.environ[\'PYTHONHASHSEED\'] = str(seed)  # Python general',
                                               'np.random.seed(seed)  # Numpy random',
                                               'random.seed(seed)  # Python random']

        # Verify that system-intelligence and conda environment logging are intact
        expected_lines_sys_intell_conda_env = [
            'def log_sys_intel_conda_env(cls):',
            'reports_output_dir = tempfile.mkdtemp()',
            'def log_system_intelligence(reports_output_dir: str):',
            'def log_conda_environment(reports_output_dir: str):',
            'query_and_export(query_scope={\'all\'},',
            'mlflow.log_artifacts(reports_output_dir, artifact_path=\'reports\')',
            f'subprocess.call([\'conda\', \'env\', \'export\', \'--name\', \'{self.project_slug_no_hyphen}\'], stdout=conda_env_filehandler)',
            f'mlflow.log_artifact(f\'{{reports_output_dir}}/{self.project_slug_no_hyphen}_conda_environment.yml\', artifact_path=\'reports\')'
        ]

        for expected_line in expected_lines_general_random_seeds + expected_lines_sys_intell_conda_env:
            if expected_line.strip() not in mlf_core_py_content:
                self.failed.append(('mlflow-general-8', f'{expected_line} not found in mlf_core.py'))

    def _print_results(self):
        console.print()
        console.rule("[bold blue]LINT RESULTS")
        console.print()
        console.print(
            f'     [bold green][[\u2714]] {len(self.passed):>4} tests passed\n     [bold yellow][[!]] {len(self.warned):>4} tests had warnings\n'
            f'     [bold red][[\u2717]] {len(self.failed):>4} tests failed',
            overflow="ellipsis",
            highlight=False
        )

        # Helper function to format test links nicely
        def format_result(test_results):
            """
            Given an list of error message IDs and the message texts, return a nicely formatted
            string for the terminal with appropriate ASCII colours.
            """
            results = []
            for eid, msg in test_results:
                results.append(f"1. [https://mlf-core.readthedocs.io/en/latest/lint.html#{eid}]"
                               f"(https://mlf-core.readthedocs.io/en/latest/lint.html#{eid}) : {msg}")
            return rich.markdown.Markdown("\n".join(results))

        if len(self.passed) > 0:
            console.print()
            console.rule("[bold green][[\u2714]] Tests Passed", style="green")
            console.print(rich.panel.Panel(format_result(self.passed), style="green"), overflow="ellipsis")
        if len(self.warned) > 0:
            console.print()
            console.rule("[bold yellow][[!]] Test Warnings", style="yellow")
            console.print(rich.panel.Panel(format_result(self.warned), style="yellow"), overflow="ellipsis")
        if len(self.failed) > 0:
            console.print()
            console.rule("[bold red][[\u2717]] Test Failures", style="red")
            console.print(rich.panel.Panel(format_result(self.failed), style="red"), overflow="ellipsis")

    def _wrap_quotes(self, files):
        if not isinstance(files, list):
            files = [files]
        bfiles = [f"`{file}`" for file in files]

        return " or ".join(bfiles)

    def _strip_ansi_codes(self, string, replace_with=""):
        # https://stackoverflow.com/a/14693789/713980
        ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")

        return ansi_escape.sub(replace_with, string)


def files_exist_linting(self,
                        files_fail: list,
                        files_fail_ifexists: list,
                        files_warn: list,
                        files_warn_ifexists: list,
                        is_subclass_calling=True,
                        handle: str = 'general') -> None:
    """
    Verifies that passed lists of files exist or do not exist.
    Depending on the desired result passing, warning or failing results are appended to the linter object.

    :param self: Linter object
    :param files_fail: list of files which must exist or linting will fail
    :param files_fail_ifexists: list of files which are not allowed to exist or linting will fail
    :param files_warn: list of files which should exist or linting will warn
    :param files_warn_ifexists: list of files which should exist or linting will warn
    :param is_subclass_calling: indicates whether the subclass of TemplateLinter called the linting (specific) or itw as the general linting
    :param handle: template handle or general if not template specific
    """
    # Files that cause an error if they don't exist
    all_exists = True
    for files in files_fail:
        if not any([os.path.isfile(pf(self, f)) for f in files]):
            all_exists = False
            self.failed.append((f'{handle}-1', f'File not found: {self._wrap_quotes(files)}'))
    # flag that indicates whether all required files exist or not
    if all_exists:
        # called linting from a specific template linter
        if is_subclass_calling:
            self.passed.append((f'{handle}-1', f'All required {handle} specific files were found!'))
        # called as general linting
        else:
            self.passed.append((f'{handle}-1', f'All required {handle} files were found!'))

    # Files that cause a warning if they don't exist
    for files in files_warn:
        if any([os.path.isfile(pf(self, f)) for f in files]):
            # pass cause if a file was found it will be summarised in one "all required files found" statement
            pass
        else:
            self.warned.append((f'{handle}-1', f'File not found: {self._wrap_quotes(files)}'))

    # Files that cause an error if they exist
    for file in files_fail_ifexists:
        if os.path.isfile(pf(self, file)):
            self.failed.append((f'{handle}-1', f'File must be removed: {self._wrap_quotes(file)}'))

    # Files that cause a warning if they exist
    for file in files_warn_ifexists:
        if os.path.isfile(pf(self, file)):
            self.warned.append((f'{handle}-1', f'File should be removed: {self._wrap_quotes(file)}'))


class GetLintingFunctionsMeta(type):
    def get_linting_functions(cls):
        """
        Fetches all specific linting methods and returns them as a list.
        Used for all template specific linters

        :param cls: The specific linting class
        """
        specific_linter_function_names = ([func for func in dir(cls) if (callable(getattr(cls, func)) and not func.startswith('__'))])
        general_linter_function_names = ([func for func in dir(TemplateLinter) if (callable(getattr(TemplateLinter, func)) and not func.startswith('__'))])
        cls_only_funcs = list(set(specific_linter_function_names) - set(general_linter_function_names))
        cls_only_funcs.remove('lint')  # remove 'lint', since we only want the newly defined methods and not the method itself

        return cls_only_funcs

    def __call__(self, *args, **kwargs):
        # create the new class as normal
        cls = type.__call__(self, *args)

        # set the methods attribute to a list of all specific linting functions
        setattr(cls, 'methods', self.get_linting_functions())

        return cls


class ChangelogLinter:
    """
    A linter for our templates Changelog.rst to ensure consistency across templates

    :attribute self.changelog_path: Path to the changelog file
    :attribute self.main_linter: The calling linter
    :attribute self.changelog_content: The content line by line of the CHANGELOG.rst file
    :attribute self.line_counter: A counter to keep track of the currents line index
    :attribute self.header_offset: An offset value to indicate where the CHANGELOG header ends
    """

    def __init__(self, path, linter: TemplateLinter):
        """
        :param path: Path to the Changelog file
        :param linter: a reference to the calling linter
        """
        self.changelog_path = path
        self.main_linter = linter
        self.changelog_content = self.init_changelog_lint_content()
        self.line_counter = 0
        self.header_offset = 0

    def lint_header(self) -> (int, bool, bool):
        """
        Lint the header which consists of an optional label, the headline CHANGELOG and an optional small description
        """
        header_detected = False
        for line in self.changelog_content:
            # lint the header until we found a section header
            if self.match_section_header(line):
                if self.changelog_content[self.line_counter + 1] >= f'{"-" * (len(line) - 1)}\n':
                    return self.header_offset, header_detected, True
                else:
                    """
                    This also helps with bump-version and automatic section adding to ensure a correct section start.
                    Example:
                    1.2.3 (12.12.2020)
                    Some text were underscores belong for correct underline
                    """
                    self.main_linter.failed.append(('general-6', 'Invalid section header start detected!'))
                    return -1, header_detected, False
            # lint header (optional label, title and an optional small description)
            elif any(cl in line for cl in ['CHANGELOG', 'Changelog']):
                head_liner = f'{"=" * len(line)}\n'
                header_ok = self.changelog_content[self.line_counter - 1] == head_liner and self.changelog_content[self.line_counter + 1] == head_liner
                """
                Example:
                ======
                MY HEADER IS TOO LONG
                =====================
                """
                if not header_ok:
                    self.main_linter.failed.append(('general-6', 'Your Changelog header syntax does not match length of your Changelogs title!'))
                    return -1, header_detected, False
                header_detected = True

            if self.header_offset >= len(self.changelog_content) - 2:
                """
                A Changelog EVER should contain at least one section (thus when the template has been created)!
                Example
                .. changelog_f

                =========
                CHANGELOG
                =========

                End
                """
                self.main_linter.failed.append(('general-6', 'No changelog sections detected!'))
                return -1, header_detected, False
            self.header_offset += 1
            self.line_counter += 1

    def lint_changelog_section(self) -> bool:
        """
        Lint a changelog section
        Example:
        1.2.3 (2020-12-06)
        ------------------
        **Added**
        We added ...

        **Fixed**
        We fixed ...

        **Dependencies**
        Dependencies ...

        **Deprecated**
        Whats deprecated now ...
        """
        # this will actually not generate a copy of the list, it just copies a reference
        section_changelog_content = self.changelog_content[self.header_offset:]
        versions = []

        # define separator keys
        def split_condition(x):
            is_split_key = self.match_section_header(x)
            # keep track of the versions
            if is_split_key:
                versions.append(x)
            return is_split_key

        grouper = groupby(section_changelog_content, key=split_condition)
        sections = list((list(sect) for split, sect in grouper if not split))

        # keep track how many sections seen so far
        section_nr = 0
        # keep track of the last version of the processed section; init with high number
        last_version = '1000000.1000000.1000000'

        for section in sections:
            # check if newer sections have a strict greater version than older sections
            current_section_version = versions[section_nr][:-1].replace('-SNAPSHOT', '').split(' ')[0]
            if version.parse(current_section_version) >= version.parse(last_version):
                self.main_linter.failed.append(('general-6', 'Older sections cannot have greater version numbers than newer sections!'))
                return False
            else:
                last_version = current_section_version

            # check if ever section subheader is underlined correctly
            if not section[0] >= f'{"-" * (len(versions[section_nr]) - 1)}\n':
                self.main_linter.failed.append(('general-6', 'Your sections subheader underline does not match the headers length!'))
                return False
            try:
                index_order_ok = section.index('**Added**\n') < section.index('**Fixed**\n') < section.index('**Dependencies**\n') < \
                                 section.index('**Deprecated**\n')
                if not index_order_ok:
                    """
                    Example (for a section)
                    1.2.3 (2020-12-06)
                    **Added**

                    **Fixed**

                    **Deprecated

                    **Dependencies**

                    Dependencies and Deprecated should be changed
                    """
                    self.main_linter.failed.append(('general-6', 'Sections subheader order should be **Added**\n**Fixed**\n'
                                                                 '**Dependencies**\n**Deprecated**!'))
                    return False

            except ValueError:
                """
                Example when missing a subsection
                1.2.3 (2020-12-06)
                **Added**

                **Deprecated

                **Dependencies**

                Fixed section is missing
                """
                self.main_linter.failed.append(('general-6', 'Section misses one or more required subsections!'))
                return False
            section_nr += 1
        return True

    def match_section_header(self, line: str) -> bool:
        """
        Check if beginning of a section header was reached
        :param line: The current line that has been read
        """
        no_snapshot_header = re.compile(r'^(?<!\.)\d+(?:\.\d+){2}(?!\.) \(\d\d\d\d-\d\d-\d\d\)$')
        snapshot_header = re.compile(r'^(?<!\.)\d+(?:\.\d+){2}(?!\.)-SNAPSHOT \(\d\d\d\d-\d\d-\d\d\)$')

        return bool(no_snapshot_header.match(line)) or bool(snapshot_header.match(line))

    def init_changelog_lint_content(self) -> list:
        """
        Reads from the project's CHANGELOG.rst.
        :return: Changelog.rst contents line by line as a list
        """
        with open(self.changelog_path, 'r') as f:
            content = f.readlines()

        return content



import sys
import logging
from pathlib import Path
from ruamel.yaml import YAML
from rich import print

from mlf_core.lint.domains.mlflow import MlflowPytorchLint, MlflowTensorflowLint, MlflowXGBoostLint
from mlf_core.lint.domains.package import PackagePredictionLint
from mlf_core.lint.template_linter import TemplateLinter

log = logging.getLogger(__name__)


def lint_project(project_dir: str) -> TemplateLinter:
    """
    Verifies the integrity of a project to best coding and practices.
    Runs a set of general linting functions, which all templates share and afterwards runs template specific linting functions.
    All results are collected and presented to the user.
    """
    # Detect which template the project is based on
    template_handle = get_template_handle(project_dir)
    log.debug(f'Detected handle {template_handle}')

    switcher = {
        'mlflow-pytorch': MlflowPytorchLint,
        'mlflow-tensorflow': MlflowTensorflowLint,
        'mlflow-xgboost': MlflowXGBoostLint,
        'package-prediction': PackagePredictionLint
    }

    try:
        lint_obj = switcher.get(template_handle)(project_dir)
    except TypeError:
        print(f'[bold red]Unable to find linter for handle {template_handle}! Aborting...')
        sys.exit(1)

    # Run the linting tests
    try:
        # Run non project specific linting
        log.debug('Running general linting.')
        print('[bold blue]Running general linting')
        lint_obj.lint_project(super(lint_obj.__class__, lint_obj), is_subclass_calling=False)

        # Run the project specific linting
        log.debug(f'Running linting of {template_handle}')
        print(f'[bold blue]Running {template_handle} linting')

        lint_obj.lint()
    except AssertionError as e:
        print(f'[bold red]Critical error: {e}')
        print('[bold red] Stopping tests...')

        return lint_obj

    # Print the results
    lint_obj._print_results()

    # Exit code
    if len(lint_obj.failed) > 0:
        print(f'[bold red] {len(lint_obj.failed)} tests failed! Exiting with non-zero error code.')
        sys.exit(1)


def get_template_handle(dot_mlf_core_path: str = '.mlf_core.yml') -> str:
    """
    Reads the .mlf_core file and extracts the template handle
    :param dot_mlf_core_path: path to the .mlf_core file
    :return: found template handle
    """
    path = Path(f'{dot_mlf_core_path}/.mlf_core.yml')
    if not path.exists():
        print('[bold red].mlf_core.yml not found. Is this a mlf-core project?')
        sys.exit(1)
    yaml = YAML(typ='safe')
    dot_mlf_core_content = yaml.load(path)

    return dot_mlf_core_content['template_handle']



import os
import logging
import sys
from rich import print
from rich.style import Style
from rich.console import Console
from rich.table import Table
from rich.box import HEAVY_HEAD
from mlf_core.common.levensthein_dist import most_similar_command
from mlf_core.common.load_yaml import load_yaml_file
from mlf_core.util.dict_util import is_nested_dictionary
from mlf_core.common.suggest_similar_commands import load_available_handles

log = logging.getLogger(__name__)


class TemplateInfo:
    """
    Present info in a nice layout about specific subsets of templates
    """

    def __init__(self):
        self.WD = os.path.dirname(__file__)
        self.TEMPLATES_PATH = f'{self.WD}/../create/templates'
        self.available_handles = ''
        self.most_sim = []
        self.action = ''

    def show_info(self, handle: str) -> None:
        """
        Displays detailed information of a domain/language/template

        :param handle: domain/language/template handle (examples: cli or cli-python)
        """
        # list of all templates that should be printed according to the passed handle
        templates_to_print = []
        available_templates = load_yaml_file(f'{self.TEMPLATES_PATH}/available_templates.yml')
        specifiers = handle.split('-')
        domain = specifiers[0]
        global template_info

        # only domain OR language specified
        if len(specifiers) == 1:
            log.debug('Only domain or language was specified.')
            try:
                template_info = available_templates[domain]
            except KeyError:
                self.handle_domain_or_language_only(handle, available_templates)
        # domain, subdomain, language
        elif len(specifiers) > 2:
            log.debug('A domain, subdomain and language was specified.')
            try:
                sub_domain = specifiers[1]
                language = specifiers[2]
                template_info = available_templates[domain][sub_domain][language]
            except KeyError:
                self.handle_non_existing_command(handle, True)
        # domain, language OR domain, subdomain
        else:
            log.debug('A domain and language OR domain and a subdomain was specified.')
            try:
                second_specifier = specifiers[1]
                template_info = available_templates[domain][second_specifier]
            except KeyError:
                self.handle_non_existing_command(handle, True)

        # Add all templates under template_info to list and output them
        self.flatten_nested_dict(template_info, templates_to_print)
        self.output_table(templates_to_print, handle)

    def handle_domain_or_language_only(self, handle: str, available_templates: dict) -> None:
        """
        Try to find a similar domain or treat handle as possible language
        :param handle: The handle inputted by the user
        :param available_templates: All available templates load from yml file as dict
        """
        # try to find a similar domain
        self.handle_non_existing_command(handle)
        # we found a similar handle matching a domain, use it (suggest it maybe later)
        if self.most_sim and self.action not in ('suggest', ''):
            self.print_console_output(handle)

        # input may be a language so try this
        templates_flatted = []
        self.flatten_nested_dict(available_templates, templates_flatted)
        # load all available languages
        available_languages = self.load_available_languages(templates_flatted)
        # if handle exists as language in mlf-core output its available templates and exit with zero status
        if handle in available_languages:
            templates_to_print_ = [template for template in templates_flatted if handle in template[1]]
            self.output_table(templates_to_print_, handle)

        # the handle does not match any domain/language; is there a similar language?
        else:
            # save domain values for later use
            domain_handle, domain_action = self.most_sim, self.action
            # try to find a similar language for instant use
            self.most_sim, self.action = most_similar_command(handle, available_languages)
            # use language if available
            if self.most_sim and self.action not in ('', 'suggest'):
                self.print_console_output(handle)
            # try to suggest a similar domain handle
            elif domain_action == 'suggest' and domain_handle:
                self.print_suggestion(handle, domain_handle)
            # try to suggest a similar language handle
            elif self.action == 'suggest' and self.most_sim:
                self.print_suggestion(handle, self.most_sim)
            # we're done there is no similar handle
            else:
                self.non_existing_handle()
        sys.exit(0)

    def output_table(self, templates_to_print: list, handle: str) -> None:
        """
        Output a nice looking, rich rendered table.
        :param templates_to_print: The templates tht should go into the table
        :param handle: The handle the user inputted
        """
        for template in templates_to_print:
            template[2] = TemplateInfo.set_linebreaks(template[2])

        table = Table(title=f'[bold]Info on mlf-core´s {handle}', title_style="blue", header_style=Style(color="blue", bold=True), box=HEAVY_HEAD)
        log.debug('Building info table.')
        table.add_column("Name", justify="left", style="green", no_wrap=True)
        table.add_column("Handle", justify="left")
        table.add_column("Long Description", justify="left")
        table.add_column("Available Libraries", justify="left")
        table.add_column("Version", justify="left")

        for template in templates_to_print:
            table.add_row(f'[bold]{template[0]}', template[1], template[2], template[3], template[4])

        console = Console()
        console.print(table)

    def handle_non_existing_command(self, handle: str, run_f: bool = False) -> None:
        """
        Handle the case, when an unknown handle was entered and try to find a similar handle.
        :param handle: The non existing handle
        :param run_f: Flag that indicates whether to run print to output or not (do not run in case of languages)
        """
        self.available_handles = load_available_handles()
        self.most_sim, self.action = most_similar_command(handle, self.available_handles)
        if run_f:
            self.print_console_output(handle)

    def print_console_output(self, handle) -> None:
        """
        Print similar command output to console.
        :param handle: The handle inputted by the user
        """
        if self.most_sim:
            # found exactly one similar handle
            if len(self.most_sim) == 1 and self.action == 'use':
                print(f'[bold red]Unknown handle \'{handle}\'. See [green]mlf-core list [red]for all valids handles')
                print(f'[bold blue]Will use best match [green]{self.most_sim[0]}.\n')
                # use best match if exactly one similar handle was found
                self.show_info(self.most_sim[0])
            elif len(self.most_sim) == 1 and self.action == 'suggest':
                self.print_suggestion(handle, self.most_sim)
            else:
                # found multiple similar handles
                nl = '\n'
                print(f'[green red]Unknown handle \'{handle}\'. See [green] mlf-core list [red]for all valid handles.' +
                      f'\nMost similar handles are: [green]{nl}{nl.join(sorted(self.most_sim))}')
            sys.exit(0)

        else:
            # found no similar handles
            self.non_existing_handle()

    def print_suggestion(self, handle: str, domain_handle: list) -> None:
        """
        Output a similar command suggestion message for the user
        :param handle: Handle inputted by the user
        :param domain_handle: A list of similar commands (will contain only one element most of the time)
        """
        print(f'[bold red]Unknown handle \'{handle}\'. See [green] mlf-core list [red] for all valid handles.\n')
        print(f'[bold red] Did you mean [green]{domain_handle[0]}?\n')

    def non_existing_handle(self) -> None:
        """
        Handling key not found access error for non existing template handles.
        Displays an error message and terminates mlf-core.
        """
        print('[bold red]Handle does not exist. Please enter a valid handle.\nUse [green] mlf-core list [red]to display all template handles.')
        sys.exit(0)

    def flatten_nested_dict(self, template_info_, templates_to_print) -> None:
        """
        Flatten an arbitrarily deep nested dict and creates a list of list containing all available
        templates for the specified doamin/subdomain and/or language.

        :param template_info_: The dict containing the yaml parsed info for all available templates the user wants to gather some information
        :param templates_to_print: A list of templates string representations, that will be printed to console
        """
        if is_nested_dictionary(template_info_):
            for templ in template_info_.values():
                if not is_nested_dictionary(templ):
                    templates_to_print.append([templ['name'], templ['handle'], templ['long description'],
                                               templ['available libraries'], templ['version']])
                else:
                    self.flatten_nested_dict(templ, templates_to_print)
        else:
            # a single template to append was reached
            templates_to_print.append([template_info_['name'], template_info_['handle'], template_info_['long description'],
                                       template_info_['available libraries'], template_info_['version']])

    @staticmethod
    def set_linebreaks(desc: str) -> str:
        """
        Sets newlines after max 45 characters (or the latest space to avoid non-sense separation)

        :param desc: The parsed long description for the sepcific template
        :return: The formatted string with inserted newlines
        """
        linebreak_limit = 50
        last_space = -1
        cnt = 0
        idx = 0

        while idx < len(desc):
            if cnt == linebreak_limit:
                # set a line break at the last space encountered to avoid separating words
                desc = desc[:last_space] + '\n' + desc[last_space + 1:]
                cnt = 0
            elif desc[idx] == ' ':
                last_space = idx
            cnt += 1
            idx += 1

        return desc

    def load_available_languages(self, ls: list) -> set:
        """
        Load all available languages mlf-core supports.
        NOTE: ASSUME ALL HANDLES ARE TWO OR THREE PARTS!
        :param ls: The flattended list from the available templates dict
        :return: All available languages as a set
        """
        return {ls[1].split('-')[1] if len(ls[1].split('-')) == 2 else ls[1].split('-')[2] for ls in ls}



import os
import logging
from rich.style import Style
from rich.console import Console
from rich.table import Table
from rich.box import HEAVY_HEAD
from rich import print

from mlf_core.util.dict_util import is_nested_dictionary
from mlf_core.common.load_yaml import load_yaml_file

log = logging.getLogger(__name__)


class TemplateLister:
    """
    A class responsible for listing all available mlf-core templates in a nice layout
    """
    WD = os.path.dirname(__file__)
    TEMPLATES_PATH = f'{WD}/../create/templates'

    @classmethod
    def list_available_templates(cls) -> None:
        """
        Displays all available templates to stdout in nicely formatted yaml format.
        Omits long descriptions.
        """
        log.debug(f'Reading available_templates.yml at {cls.TEMPLATES_PATH}/available_templates.yml')
        available_templates = load_yaml_file(f'{cls.TEMPLATES_PATH}/available_templates.yml')
        print('[bold blue]Run [green]mlf-core info [blue]for long descriptions of your template of interest')
        print()

        # What we want to have are lists like
        # [['name', 'handle', 'short description', 'available libraries', 'version'], ['name', 'handle', 'short description', 'available libraries', 'version']]
        log.debug('Building list table.')
        templates_to_tabulate = []
        for language in available_templates.values():
            for val in language.values():
                # has a subdomain -> traverse dictionary a level deeper
                if is_nested_dictionary(val):
                    for val_2 in val.values():
                        templates_to_tabulate.append([
                            val_2['name'], val_2['handle'], val_2['short description'], val_2['available libraries'], val_2['version']
                        ])
                else:
                    templates_to_tabulate.append([
                        val['name'], val['handle'], val['short description'], val['available libraries'], val['version']
                    ])

        table = Table(title="[bold]All available mlf-core templates", title_style="blue", header_style=Style(color="blue", bold=True), box=HEAVY_HEAD)

        table.add_column("Name", justify="left", style="green", no_wrap=True)
        table.add_column("Handle", justify="left")
        table.add_column("Short Description", justify="left")
        table.add_column("Available Libraries", justify="left")
        table.add_column("Version", justify="left")

        for template in templates_to_tabulate:
            table.add_row(f'[bold]{template[0]}', template[1], template[2], template[3], template[4])

        console = Console()
        log.debug('Printing list table.')
        console.print(table)



import os
from pathlib import Path
from dataclasses import dataclass

from mlf_core.create.github_support import prompt_github_repo
from mlf_core.create.template_creator import TemplateCreator
from mlf_core.create.domains.mlf_core_template_struct import MlfcoreTemplateStruct
from mlf_core.custom_cli.questionary import mlf_core_questionary_or_dot_mlf_core
from mlf_core.common.version import load_mlf_core_template_version


@dataclass
class TemplateStructPackage(MlfcoreTemplateStruct):
    """
    PACKAGE-PREDICTION
    """
    framework: str = ''  # which machine learning framework to use


class PackageCreator(TemplateCreator):

    def __init__(self):
        self.package_struct = TemplateStructPackage(domain='package')
        super().__init__(self.package_struct)
        self.WD_Path = Path(os.path.dirname(__file__))
        self.TEMPLATES_PACKAGE_PATH = f'{self.WD_Path.parent}/templates/package'

        '"" TEMPLATE VERSIONS ""'
        self.PACKAGE_PREDICTION_TEMPLATE_VERSION = load_mlf_core_template_version('package-prediction', self.AVAILABLE_TEMPLATES_PATH)

    def create_template(self, path: Path, dot_mlf_core: dict or None):
        """
        Handles the package domain. Prompts the user for the language, general and domain specific options.
        """

        self.package_struct.language = mlf_core_questionary_or_dot_mlf_core(function='select',
                                                                            question='Choose the project\'s primary purpose',
                                                                            choices=['prediction'],
                                                                            default='prediction',
                                                                            dot_mlf_core=dot_mlf_core,
                                                                            to_get_property='language')

        # prompt the user to fetch general template configurations
        super().prompt_general_template_configuration(dot_mlf_core)

        # switch case statement to prompt the user to fetch template specific configurations
        switcher = {
            'prediction': self.package_prediction_options,
        }
        switcher.get(self.package_struct.language)(dot_mlf_core)

        self.package_struct.is_github_repo, \
            self.package_struct.is_repo_private, \
            self.package_struct.is_github_orga, \
            self.package_struct.github_orga \
            = prompt_github_repo(dot_mlf_core)

        if self.package_struct.is_github_orga:
            self.package_struct.github_username = self.package_struct.github_orga
        # create the chosen and configured template
        super().create_template_without_subdomain(self.TEMPLATES_PACKAGE_PATH)

        # switch case statement to fetch the template version
        switcher_version = {
            'package': self.PACKAGE_PREDICTION_TEMPLATE_VERSION,
        }
        self.package_struct.template_version, self.package_struct.template_handle = switcher_version.get(
            self.package_struct.language), f'package-{self.package_struct.language.lower()}'

        # perform general operations like creating a GitHub repository and general linting
        super().process_common_operations(path=Path(path).resolve(), domain='package', language=self.package_struct.language, dot_mlf_core=dot_mlf_core)

    def package_prediction_options(self, dot_mlf_core: dict or None):
        """ Prompts for package-prediction specific options and saves them into the PackageTemplateStruct """
        self.package_struct.framework = mlf_core_questionary_or_dot_mlf_core(function='select',
                                                                             question='Choose a framework',
                                                                             choices=['pytorch', 'tensorflow', 'xgboost'],
                                                                             default='pytorch',
                                                                             dot_mlf_core=dot_mlf_core,
                                                                             to_get_property='framework')



# -*- coding: utf-8 -*-

"""Entry point for mlf-core."""
import logging
import os
import sys
import click
from pathlib import Path

import rich.logging
from rich import traceback
from rich import print

from mlf_core.bump_version.bump_version import VersionBumper
from mlf_core.config.config import ConfigCommand
from mlf_core.create.create import choose_domain
from mlf_core.custom_cli.click import HelpErrorHandling, print_project_version, print_mlfcore_version, CustomHelpSubcommand, CustomArg
from mlf_core.info.info import TemplateInfo
from mlf_core.lint.lint import lint_project
from mlf_core.list.list import TemplateLister
from mlf_core.custom_cli.questionary import mlf_core_questionary_or_dot_mlf_core
from mlf_core.sync.sync import TemplateSync
from mlf_core.upgrade.upgrade import UpgradeCommand
from mlf_core.common.load_yaml import load_yaml_file

WD = os.path.dirname(__file__)
log = logging.getLogger()


def main():
    traceback.install(width=200, word_wrap=True)
    print(r"""[bold blue]
███    ███ ██      ███████      ██████  ██████  ██████  ███████ 
████  ████ ██      ██          ██      ██    ██ ██   ██ ██      
██ ████ ██ ██      █████ █████ ██      ██    ██ ██████  █████ 
██  ██  ██ ██      ██          ██      ██    ██ ██   ██ ██    
██      ██ ███████ ██           ██████  ██████  ██   ██ ███████ 
                                                               
        """)

    print('[bold blue]Run [green]mlf-core --help [blue]for an overview of all commands\n')

    # Is the latest mlf-core version installed? Upgrade if not!
    if not UpgradeCommand.check_mlf_core_latest():
        print('[bold blue]Run [green]mlf-core upgrade [blue]to get the latest version.')
    mlf_core_cli()


@click.group(cls=HelpErrorHandling)
@click.option('--version', is_flag=True, callback=print_mlfcore_version, expose_value=False, is_eager=True, help='Print the current mlf-core version.')
@click.option('-v', '--verbose', is_flag=True, default=False, help='Enable verbose output (print debug statements).')
@click.option("-l", "--log-file", help="Save a verbose log to a file.")
@click.pass_context
def mlf_core_cli(ctx, verbose, log_file):
    """
    Create state of the art projects from production ready templates.
    """
    # Set the base logger to output DEBUG
    log.setLevel(logging.DEBUG)

    # Set up logs to the console
    log.addHandler(
        rich.logging.RichHandler(
            level=logging.DEBUG if verbose else logging.INFO,
            console=rich.console.Console(file=sys.stderr),
            show_time=True,
            markup=True,
        )
    )

    # Set up logs to a file if we asked for one
    if log_file:
        log_fh = logging.FileHandler(log_file, encoding="utf-8")
        log_fh.setLevel(logging.DEBUG)
        log_fh.setFormatter(logging.Formatter("[%(asctime)s] %(name)-20s [%(levelname)-7s]  %(message)s"))
        log.addHandler(log_fh)


@mlf_core_cli.command(short_help='Create a new project using one of our templates.', cls=CustomHelpSubcommand)
@click.argument('path', type=click.Path(), default=Path.cwd(), helpmsg='Path where the project should be created at.', cls=CustomArg)  # type: ignore
@click.option('--domain', type=click.Choice(['cli', 'lib', 'gui', 'web', 'pub']),
              help='The projects domain with currently cli, lib, gui, web and pub supported.')
def create(path: Path, domain: str) -> None:
    """
    Create a new project using one of our templates.

    Creates a new project based on one of mlf-core's templates.
    You will be prompted for the domain, subdomain (if applicable), language and possibly framework.
    Template specific prompts follow. If you do not yet have a mlf-core config file you may be asked to create one first.
    Next, you will be asked whether you want to use mlf-core's Github support create a repository, push your template and enable a few settings.
    After the project has been created it will be linted and you will be notified of any TODOs.
    """
    choose_domain(path, domain, None)


@mlf_core_cli.command(short_help='Lint your existing mlf-core project.', cls=CustomHelpSubcommand)
@click.argument('project_dir', type=click.Path(), default=Path(str(Path.cwd())), helpmsg='Relative path to projects directory.', cls=CustomArg)
def lint(project_dir) -> None:
    """
    Lint your existing mlf-core project.

    Verify that your existing project still adheres to mlf_core's standards.
    mlf-core runs several general linting functions, which all templates share.
    Examples include a consistent project version, the existence of documentation and whether cookiecutter statements are still left.
    Afterwards, template specific linting is invoked. cli-python for example may check for the existence of a setup.py file.
    Both results are collected and displayed.
    """
    lint_project(project_dir)


@mlf_core_cli.command(short_help='List all available mlf-core templates.', cls=CustomHelpSubcommand)
def list() -> None:
    """
    List all available mlf-core templates.

    Get an overview of all existing mlf-core templates.
    The output only consists of a short description for all templates.
    To get a detailed overview of a specific subset of templates use info.
    """
    TemplateLister.list_available_templates()


@mlf_core_cli.command(short_help='Get detailed info on a mlf-core template domain or a single template.', cls=CustomHelpSubcommand)
@click.argument('handle', type=str, required=False, helpmsg='Language/domain of templates of interest.', cls=CustomArg)
@click.pass_context
def info(ctx, handle: str) -> None:
    """
    Get detailed info on a mlf_core template domain or a single template.

    list only provides an overview of all templates.
    Info provides a long description for a specific subset of templates.
    Pass a domain, language or full handle (e.g. cli-python).
    """
    if not handle:
        HelpErrorHandling.args_not_provided(ctx, 'info')
    else:
        template_info = TemplateInfo()
        template_info.show_info(handle.lower())


@mlf_core_cli.command(short_help='Sync your project with the latest template release.', cls=CustomHelpSubcommand)
@click.argument('project_dir', type=str, default=Path(f'{Path.cwd()}'),
                helpmsg='The projects top level directory you would like to sync. Default is current ''working directory.', cls=CustomArg)
@click.option('--set-token', '-st', is_flag=True, help='Set sync token to a new personal access token of the current repo owner.')
@click.argument('pat', type=str, required=False, helpmsg='Personal access token. Not needed for manual, local syncing!', cls=CustomArg)
@click.argument('username', type=str, required=False, helpmsg='Github username. Not needed for manual, local syncing!', cls=CustomArg)
@click.option('--check-update', '-ch', is_flag=True, help='Check whether a new template version is available for your project.')
def sync(project_dir, set_token, pat, username, check_update) -> None:
    """
    Sync your project with the latest template release.
    mlf_core regularly updates its templates.
    To ensure that you have the latest changes you can invoke sync, which submits a pull request to your Github repository (if existing).
    If no repository exists the TEMPLATE branch will be updated and you can merge manually.
    """
    project_dir_path = Path(project_dir).resolve()
    log.debug(f'Loading project information from .mlf_core.yml file located at {project_dir}')
    project_data = load_yaml_file(f'{project_dir}/.mlf_core.yml')
    log.debug(f'Set project top level path to given path argument {project_dir_path}')
    # if set_token flag is set, update the sync token value and exit
    if set_token:
        log.debug('Running sync to update sync token in repo.')
        try:
            if project_data['is_github_repo']:
                log.debug(f'Project is a Github repo. Using {project_data["github_username"]} as username.')
                TemplateSync.update_sync_token(project_name=project_data['project_slug'], gh_username=project_data['github_username'])
            else:
                print('[bold red]Your current project does not seem to have a Github repository!')
                sys.exit(1)
        except FileNotFoundError:
            print(f'[bold red]There exists no .mlf_core.yml file at {project_dir_path}. Is this a mlf-core project?')
        except KeyError:
            print('[bold red]Your token value is not a valid personal access token for your account.')
            sys.exit(1)
        sys.exit(0)

    log.debug('Initializing syncer object.')
    syncer = TemplateSync(new_template_version='', project_dir=project_dir_path, gh_username=username, token=pat)
    # check for template version updates
    log.debug('Checking for major/minor or patch version changes in mlf-core templates.')
    major_change, minor_change, patch_change, project_template_version, mlf_core_template_version = syncer.has_template_version_changed(project_dir_path)
    syncer.new_template_version = mlf_core_template_version
    # check for user without actually syncing
    if check_update:
        log.debug('Running snyc to manually check whether a new template version is available.')
        # a template update has been released by mlf-core
        if any(change for change in (major_change, minor_change, patch_change)):
            print(f'[bold blue]Your templates version received an update from {project_template_version} to {mlf_core_template_version}!\n'
                  f' Use [green]mlf-core sync [blue]to sync your project')
        else:
            print('[bold blue]Using the latest template version. No sync required.')
        # exit without syncing
        sys.exit(0)

    # set sync flags indicating a major, minor or patch update
    syncer.major_update = major_change
    syncer.minor_update = minor_change
    syncer.patch_update = patch_change
    log.debug('Major template update found.' if major_change else 'Minor template update found.' if minor_change else 'Patch template update found.' if
              patch_change else 'No template update found.')

    # sync the project if any changes were detected
    if any(change for change in (major_change, minor_change, patch_change)):
        # If required sync level is set and sync is enabled -> sync
        if syncer.should_run_sync():
            log.debug('Starting sync.')
            syncer.sync()
        else:
            print('[bold red]Aborting sync due to set level constraints or sync being disabled. '
                  'You can set the level any time in your mlf_core.cfg in the sync_level section and sync again.')
    else:
        print('[bold blue]No changes detected. Your template is up to date.')


@mlf_core_cli.command('bump-version', short_help='Bump the version of an existing mlf-core project.', cls=CustomHelpSubcommand)
@click.argument('new_version', type=str, required=False, helpmsg='New project version in a valid format.', cls=CustomArg)
@click.argument('project-dir', type=click.Path(), default=Path(f'{Path.cwd()}'), helpmsg='Relative path to the projects directory.', cls=CustomArg)
@click.option('--downgrade', '-d', is_flag=True, help='Set this flag to downgrade a version.')
@click.option('--project-version', is_flag=True, callback=print_project_version, expose_value=False, is_eager=True, help='Print your projects version and exit')
@click.pass_context
def bump_version(ctx, new_version, project_dir, downgrade) -> None:
    """
    Bump the version of an existing mlf-core project.

    INFO on valid versions: All versions must match the format like 1.0.0 or 1.1.0-SNAPSHOT; these are the only valid
    version formats mlf-core allows. A valid version therefore contains a three digits (in the range from 0 to however large it will grow)
    separated by two dots.
    Optional is the -SNAPSHOT at the end (for JVM templates especially). NOTE that versions like 1.2.3.4 or 1.2 WILL NOT be recognized as valid versions as
    well as no substring of them will be recognized.

    Unless the user uses downgrade mode via the -d flag, a downgrade of a version is never allowed. Note that bump-version with the new version
    equals the current version is never allowed, either with or without -d.
    """
    if not new_version:
        HelpErrorHandling.args_not_provided(ctx, 'bump-version')
    else:
        # if the path entered ends with a trailing slash remove it for consistent output
        if str(project_dir).endswith('/'):
            project_dir = Path(str(project_dir).replace(str(project_dir)[len(str(project_dir)) - 1:], ''))

        version_bumper = VersionBumper(project_dir, downgrade)
        # lint before run bump-version
        version_bumper.lint_before_bump()
        # only run bump-version if conditions are met
        if version_bumper.can_run_bump_version(new_version, project_dir):
            # only run "sanity" checker when the downgrade flag is not set
            if not downgrade:
                # if the check fails, ask the user for confirmation
                if version_bumper.check_bump_range(version_bumper.CURRENT_VERSION.split('-')[0], new_version.split('-')[0]):
                    version_bumper.bump_template_version(new_version, project_dir)
                elif mlf_core_questionary_or_dot_mlf_core(function='confirm',
                                                          question=f'Bumping from {version_bumper.CURRENT_VERSION} to {new_version} seems not reasonable.\n'
                                                                   f'Do you really want to bump the project version?',
                                                          default='n'):
                    print('\n')
                    version_bumper.bump_template_version(new_version, project_dir)
            else:
                version_bumper.bump_template_version(new_version, project_dir)
        else:
            sys.exit(1)


@mlf_core_cli.command(short_help='Configure your general settings and github credentials.', cls=CustomHelpSubcommand)
@click.option('--view', '-v', is_flag=True, help='View the current mlf-core configuration.')
@click.argument('section', type=str, required=False, helpmsg='Section to configure (all, general or pat)', cls=CustomArg)
@click.pass_context
def config(ctx, view: bool, section: str) -> None:
    """
    Configure your general settings and Github credentials for reuse.
    Available options (sections) are:

    \b
    - general: set your fullname, email and Github username
    - pat: set your Github personal access token for Github repository creation
    - all: calls general and pat
    """
    if view:
        ConfigCommand.view_current_config()
        sys.exit(0)
    if section == 'general':
        # set the full_name and email for reuse in the creation process
        ConfigCommand.config_general_settings()
    elif section == 'pat':
        # set github username and encrypted personal access token
        ConfigCommand.config_pat()
    elif section == 'all':
        # set everything
        ConfigCommand.all_settings()
        # empty section argument causes a customized error
    elif not section:
        HelpErrorHandling.args_not_provided(ctx, 'config')
        # check if a similar section handle can be used/suggested
    else:
        ConfigCommand.similar_handle(section)


@mlf_core_cli.command(short_help='Fix artifact location path for local all mlruns.', cls=CustomHelpSubcommand)
@click.argument('path', type=str, default=os.getcwd(), required=False, helpmsg='Path to the root of the mlruns folder.', cls=CustomArg)
@click.pass_context
def fix_artifact_paths(ctx, path: str) -> None:
    """
    Ensures that the paths of all locally saved MLflow artifacts are fixed to display them on the current machine.
    """
    for meta_yaml in Path(f'{path}/mlruns').rglob('meta.yaml'):
        with open(meta_yaml.absolute()) as meta_yaml_file:
            content = meta_yaml_file.readlines()
            if 'file://' not in content[0]:
                print(f'[bold yellow]Skipping path fixing for: {meta_yaml.absolute()}. Run was not saved locally.')
        print(f'[bold blue]Fixing path for: {meta_yaml.absolute()}')
        with open(meta_yaml.absolute()) as meta_yaml_file:
            content = meta_yaml_file.readlines()
            if 'artifact_location' in content[0]:
                content[0] = f'artifact_location: file://{meta_yaml.absolute().__str__()[:-10]}\n'
            else:
                content[0] = f'artifact_uri: file://{meta_yaml.absolute().__str__()[:-10]}/artifacts\n'

        with open(meta_yaml.absolute(), 'w') as meta_yaml_file:
            meta_yaml_file.writelines(content)


@mlf_core_cli.command(short_help='Check for a newer version of mlf-core and upgrade if required.', cls=CustomHelpSubcommand)
def upgrade() -> None:
    """
    Checks whether the locally installed version of mlf-core is the latest.
    If not pip will be invoked to upgrade mlf-core to the latest version.
    """
    UpgradeCommand.check_upgrade_mlf_core()


if __name__ == "__main__":
    traceback.install()
    sys.exit(main())  # pragma: no cover



import logging
import sys
from typing import Union

import questionary
from prompt_toolkit.styles import Style
from rich import print

log = logging.getLogger(__name__)


mlf_core_style = Style([
    ('qmark', 'fg:#0000FF bold'),  # token in front of the question
    ('question', 'bold'),  # question text
    ('answer', 'fg:#008000 bold'),  # submitted answer text behind the question
    ('pointer', 'fg:#0000FF bold'),  # pointer used in select and checkbox prompts
    ('highlighted', 'fg:#0000FF bold'),  # pointed-at choice in select and checkbox prompts
    ('selected', 'fg:#008000'),  # style for a selected item of a checkbox
    ('separator', 'fg:#cc5454'),  # separator in lists
    ('instruction', ''),  # user instructions for select, rawselect, checkbox
    ('text', ''),  # plain text
    ('disabled', 'fg:#FF0000 italic')  # disabled choices for select and checkbox prompts
])


def mlf_core_questionary_or_dot_mlf_core(function: str,
                                         question: str,
                                         choices: list = None,
                                         default: str = None,
                                         dot_mlf_core: dict = None,
                                         to_get_property: str = None) -> Union[str, bool]:
    """
    Custom selection based on Questionary. Handles keyboard interrupts and default values.

    :param function: The function of questionary to call (e.g. select or text). See https://github.com/tmbo/questionary for all available functions.
    :param choices: List of all possible choices.
    :param question: The question to prompt for. Should not include default values or colons.
    :param default: A set default value, which will be chosen if the user does not enter anything.
    :param dot_mlf_core: A dictionary, which contains the whole .mlf_core.yml content
    :param to_get_property: A key, which must be in the dot_mlf_core file, which is used to fetch the read in value from the .mlf_core.yml file
    :return: The chosen answer.
    """
    # First check whether a dot_mlf_core was passed and whether it contains the desired property -> return it if so
    try:
        if dot_mlf_core:
            if to_get_property in dot_mlf_core:
                return dot_mlf_core[to_get_property]
    except KeyError:
        log.debug(f'.mlf_core.yml file was passed when creating a project, but key {to_get_property}'
                  f' does not exist in the dot_mlf_core dictionary! Assigning default {default} to {to_get_property}.')
        return default

    # There is no .mlf_core.yml file aka dot_mlf_core dict passed -> ask for the properties
    answer = ''
    try:
        if function == 'select':
            if default not in choices:
                log.debug(f'Default value {default} is not in the set of choices!')
            answer = getattr(questionary, function)(f'{question}: ', choices=choices, style=mlf_core_style).unsafe_ask()
        elif function == 'password':
            while not answer or answer == '':
                answer = getattr(questionary, function)(f'{question}: ', style=mlf_core_style).unsafe_ask()
        elif function == 'text':
            if not default:
                log.debug('Tried to utilize default value in questionary prompt, but is None! Please set a default value.')
                default = ''
            answer = getattr(questionary, function)(f'{question} [{default}]: ', style=mlf_core_style).unsafe_ask()
        elif function == 'confirm':
            default_value_bool = True if default == 'Yes' or default == 'yes' else False
            answer = getattr(questionary, function)(f'{question} [{default}]: ', style=mlf_core_style, default=default_value_bool).unsafe_ask()
        else:
            log.debug(f'Unsupported questionary function {function} used!')

    except KeyboardInterrupt:
        print('[bold red] Aborted!')
        sys.exit(1)
    if answer is None or answer == '':
        answer = default

    return answer



from rich import print


def fix_short_title_underline(path_to_rst_file: str) -> None:
    """
    Fixes too short underlines of titles of *.rst files
    example:
    MLF-CORE
    ======
    ->
    MLF-CORE
    ============

    :param path_to_rst_file: Path to the *.rst file (usually index.rst)
    """
    print('[bold blue]Fixing too short underlines of *.rst file (usually index.rst)')
    try:
        with open(path_to_rst_file) as f:
            content = f.readlines()
            # Fix the underlined title by replacing the short underline with the correct length
            len_header = len(content[0])
            content[1] = len_header * '='
            # Write everything back
            with open(path_to_rst_file, 'w') as file:
                file.writelines(content)
    except FileNotFoundError:
        print(f'[bold yellow]Unable ot find rst file: {path_to_rst_file}')



import json
import urllib
import sys

from pkg_resources import parse_version

import mlf_core

from urllib.error import HTTPError, URLError
from subprocess import Popen, PIPE, check_call
from mlf_core.custom_cli.questionary import mlf_core_questionary_or_dot_mlf_core
from rich import print


class UpgradeCommand:
    """
    Responsible for checking for newer versions mlf-core and upgrading it if required.
    """

    @staticmethod
    def check_upgrade_mlf_core() -> None:
        """
        Checks whether the locally installed version of mlf-core is the latest.
        If not it prompts whether to upgrade and runs the upgrade command if desired.
        """
        if not UpgradeCommand.check_mlf_core_latest():
            if mlf_core_questionary_or_dot_mlf_core(function='confirm',
                                                    question='Do you want to upgrade?',
                                                    default='y'):
                UpgradeCommand.upgrade_mlf_core()

    @classmethod
    def check_mlf_core_latest(cls) -> bool:
        """
        Checks whether the locally installed version of mlf-core is the latest available on PyPi.

        :return: True if locally version is the latest or PyPI is inaccessible, false otherwise
        """
        latest_local_version = mlf_core.__version__
        sliced_local_version = latest_local_version[:-9] if latest_local_version.endswith('-SNAPSHOT') else latest_local_version
        try:
            # Retrieve info on latest version
            # Adding nosec (bandit) here, since we have a hardcoded https request
            # It is impossible to access file:// or ftp://
            # See: https://stackoverflow.com/questions/48779202/audit-url-open-for-permitted-schemes-allowing-use-of-file-or-custom-schemes
            req = urllib.request.Request('https://pypi.org/pypi/mlf-core/json')  # nosec
            with urllib.request.urlopen(req, timeout=1) as response:  # nosec
                contents = response.read()
                data = json.loads(contents)
                latest_pypi_version = data['info']['version']
        except (HTTPError, TimeoutError, URLError):
            print('[bold red]Unable to contact PyPI to check for the latest mlf-core version. Do you have an internet connection?')
            # Returning true by default, since this is not a serious issue
            return True

        if parse_version(sliced_local_version) > parse_version(latest_pypi_version):
            print(f'[bold yellow]Installed version {latest_local_version} of mlf-core is newer than the latest release {latest_pypi_version}!'
                  f' You are running a nightly version and features may break!')
        elif parse_version(sliced_local_version) == parse_version(latest_pypi_version):
            return True
        else:
            print(f'[bold red]Installed version {latest_local_version} of mlf-core is outdated. Newest version is {latest_pypi_version}!')
            return False

    @classmethod
    def upgrade_mlf_core(cls) -> None:
        """
        Calls pip as a subprocess with the --upgrade flag to upgrade mlf-core to the latest version.
        """
        if not UpgradeCommand.is_pip_accessible():
            sys.exit(1)
        try:
            check_call([sys.executable, '-m', 'pip', 'install', '--upgrade', 'mlf-core'])
        except Exception as e:
            print('[bold red]Unable to upgrade mlf-core')
            print(f'[bold red]Exception: {e}')

    @classmethod
    def is_pip_accessible(cls) -> bool:
        """
        Verifies that pip is accessible and in the PATH.

        :return: True if accessible, false if not
        """
        pip_installed = Popen(['pip', '--version'], stdout=PIPE, stderr=PIPE, universal_newlines=True)
        (git_installed_stdout, git_installed_stderr) = pip_installed.communicate()
        if pip_installed.returncode != 0:
            print('[bold red]Unable to find \'pip\' in the PATH. Is it installed?')
            print('[bold red]Run command was [green]\'pip --version \'')
            return False

        return True



from pathlib import Path

from mlf_core.create.domains.mlflow_creator import MlflowCreator
from mlf_core.create.domains.package_creator import PackageCreator
from mlf_core.custom_cli.questionary import mlf_core_questionary_or_dot_mlf_core


def choose_domain(path: Path, domain: str or None, dot_mlf_core: dict = None):
    """
    Prompts the user for the template domain.
    Creates the .mlf_core file.
    Prompts the user whether or not to create a Github repository

    :param path: The path, the project should be created at
    :param domain: Template domain
    :param dot_mlf_core: Dictionary created from the .mlf_core.yml file. None if no .mlf_core.yml file was used.
    """
    if not domain:
        domain = mlf_core_questionary_or_dot_mlf_core(function='select',
                                                      question='Choose the project\'s domain',
                                                      choices=['mlflow', 'package'],
                                                      default='mlflow',
                                                      dot_mlf_core=dot_mlf_core,
                                                      to_get_property='domain')

    switcher = {
        'mlflow': MlflowCreator,
        'package': PackageCreator
    }

    creator_obj = switcher.get(domain.lower())()
    creator_obj.create_template(path, dot_mlf_core)



from pathlib import Path
from ruamel.yaml import YAML


def load_yaml_file(yaml_file_path: str) -> dict:
    """
    Loads a yaml file and returns the content as nested dictionary.

    :return: nested dictionary as the content of the yaml file
    """
    path = Path(yaml_file_path)
    yaml = YAML()
    yaml.boolean_representation = ['False', 'True']

    return yaml.load(path)



import os
from pathlib import Path


def delete_dir_tree(directory: Path) -> None:
    """
    Recursively delete a whole directory and its content.
    Since there is no built-in function for this in the pathlib API and we want to keep it consistent, we have to use a custom function.

    :param directory: The directory that should be removed
    """

    dir = directory

    for file in dir.iterdir():
        if file.is_dir():
            delete_dir_tree(file)
        else:
            file.unlink()
    dir.rmdir()


def pf(calling_class, file_path: str) -> str:
    """

    :param calling_class: the class of which this method is called
    :param file_path: path to file
    :return: joined path
    """
    return os.path.join(calling_class.path, file_path)


def find_filepath_in_dir(file_name: str, path: str, default: str = None) -> str:
    """
    Looks for a filepath in a directory and returns it.

    :param file_name: File to look for
    :param path: Start path to start looking for the file
    :default: The default expected file path
    :return: absolute file path of the desired file
    """
    result = None
    for root, dirs, files in os.walk(path):
        for file in files:
            if file_name in file:
                return os.path.join(root, file)

    if not result:
        return default
